import numpy as np

import jesse.indicators as ta
from jesse.factories import candles_from_close_prices
import pytest
from .data.test_candles_indicators import *

matypes = 39


def test_acosc():
    candles = np.array(test_candles_19)
    single = ta.acosc(candles)
    seq = ta.acosc(candles, sequential=True)

    assert type(single).__name__ == 'AC'
    assert round(single.osc, 2) == -21.97
    assert round(single.change, 2) == -9.22

    assert seq.osc[-1] == single.osc
    assert len(seq.osc) == len(candles)


def test_ad():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.ad(candles)
    seq = ta.ad(candles, sequential=True)
    assert round(single, 0) == 6346031
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_adosc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.adosc(candles, fast_period=3, slow_period=10)
    seq = ta.adosc(candles, fast_period=3, slow_period=10, sequential=True)

    assert round(single / 1000000, 3) == -1.122
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_adx():
    candles = np.array(test_candles_10)

    single = ta.adx(candles)
    seq = ta.adx(candles, sequential=True)

    assert round(single) == 26
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_adxr():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.adxr(candles, period=14)
    seq = ta.adxr(candles, period=14, sequential=True)

    assert round(single, 0) == 29
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_alligator():
    candles = np.array(test_candles_19)
    single = ta.alligator(candles)
    seq = ta.alligator(candles, sequential=True)

    assert type(single).__name__ == 'AG'
    assert round(single.teeth, 0) == 236
    assert round(single.jaw, 0) == 233
    assert round(single.lips, 0) == 224

    assert seq.teeth[-1] == single.teeth
    assert len(seq.teeth) == len(candles)


def test_alma():
    candles = np.array(test_candles_19)
    single = ta.alma(candles)
    seq = ta.alma(candles, sequential=True)

    assert round(single, 2) == 179.17
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ao():
    candles = np.array(test_candles_19)
    single = ta.ao(candles)
    seq = ta.ao(candles, sequential=True)

    assert round(single.osc, 0) == -46
    assert len(seq[-1]) == len(candles)
    assert seq.osc[-1] == single.osc


def test_apo():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.apo(candles, fast_period=12, slow_period=26, matype=1)
    seq = ta.apo(candles, fast_period=12, slow_period=26, matype=1, sequential=True)

    assert round(single, 2) == -15.32
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_aroon():
    candles = np.array(test_candles_19)

    aroon = ta.aroon(candles, period=14)
    assert type(aroon).__name__ == 'AROON'
    assert round(aroon.down, 2) == 100
    assert round(aroon.up, 2) == 64.29

    seq_aroon = ta.aroon(candles, period=14, sequential=True)
    assert seq_aroon.down[-1] == aroon.down
    assert len(seq_aroon.down) == len(candles)
    assert len(seq_aroon.up) == len(candles)


def test_aroon_osc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.aroonosc(candles, period=14)
    seq = ta.aroonosc(candles, period=14, sequential=True)

    assert round(single, 2) == -35.71
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_atr():
    candles = np.array(test_candles_2)

    single = ta.atr(candles)
    seq = ta.atr(candles, sequential=True)

    assert round(single, 1) == 2.8
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_avgprice():
    candles = np.array(test_candles_19)

    single = ta.avgprice(candles)
    seq = ta.avgprice(candles, sequential=True)

    assert round(single, 1) == 149.8
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_bandpass():
    candles = np.array(test_candles_19)

    bp = ta.bandpass(candles)
    assert type(bp).__name__ == 'BandPass'
    assert round(bp.bp, 2) == -7.56
    assert round(bp.bp_normalized, 2) == -0.29
    assert bp.signal == 1
    assert round(bp.trigger, 2) == -0.27

    seq_bp = ta.bandpass(candles, sequential=True)
    assert seq_bp.bp[-1] == bp.bp
    assert len(seq_bp.bp) == len(candles)
    assert len(seq_bp.bp_normalized) == len(candles)
    assert len(seq_bp.signal) == len(candles)
    assert len(seq_bp.trigger) == len(candles)


def test_beta():
    # use the same candles as mama_candles
    candles = np.array(test_candles_sol)
    benchmark_candles = np.array(test_candles_sol)

    single = ta.beta(candles, benchmark_candles)
    seq = ta.beta(candles, benchmark_candles, sequential=True)

    assert round(single, 2) == 1.
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_bollinger_bands():
    candles = np.array(test_candles_11)

    bb = ta.bollinger_bands(candles)
    u, m, l = bb
    assert type(bb).__name__ == 'BollingerBands'
    assert round(u, 1) == 145.8
    assert round(m, 1) == 141.2
    assert round(l, 1) == 136.7

    seq_bb = ta.bollinger_bands(candles, sequential=True)
    assert seq_bb.upperband[-1] == u
    assert len(seq_bb.upperband) == len(candles)
    assert len(seq_bb.middleband) == len(candles)
    assert len(seq_bb.lowerband) == len(candles)


def test_bollinger_bands_width():
    candles = np.array(test_candles_12)

    single = ta.bollinger_bands_width(candles)
    seq = ta.bollinger_bands_width(candles, sequential=True)

    assert round(single, 4) == 0.0771
    assert len(seq) == len(candles)
    assert round(seq[-1], 4) == round(single, 4)


def test_bop():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.bop(candles)
    seq = ta.bop(candles, sequential=True)

    assert round(single, 2) == -0.92
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.cc(candles)
    seq = ta.cc(candles, sequential=True)

    assert round(single, 0) == -41
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cci():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.cci(candles, period=14)
    seq = ta.cci(candles, period=14, sequential=True)

    assert round(single, 2) == -285.29
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cfo():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.cfo(candles)
    seq = ta.cfo(candles, sequential=True)

    assert round(single, 2) == -66.53
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cg():
    candles = np.array(test_candles_19)
    single = ta.cg(candles)
    seq = ta.cg(candles, sequential=True)
    assert round(single, 2) == -5.37
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_chande():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single_long = ta.chande(candles)
    seq_long = ta.chande(candles, sequential=True)

    single_short = ta.chande(candles, direction="short")
    seq_short = ta.chande(candles, direction="short", sequential=True)

    assert round(single_long, 0) == 213
    assert round(single_short, 0) == 165

    assert len(seq_short) == len(candles)
    assert len(seq_long) == len(candles)
    assert seq_long[-1] == single_long
    assert seq_short[-1] == single_short


def test_chop():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.chop(candles)
    seq = ta.chop(candles, sequential=True)

    assert round(single, 2) == 28.82
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cksp():
    candles = np.array(test_candles_19)

    single = ta.cksp(candles)
    assert type(single).__name__ == 'CKSP'
    assert round(single.long, 2) == 247.62
    assert round(single.short, 2) == 127.89

    seq = ta.cksp(candles, sequential=True)
    assert seq.long[-1] == single.long
    assert seq.short[-1] == single.short
    assert len(seq.long) == len(candles)
    assert len(seq.short) == len(candles)


def test_cmo():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.cmo(candles, period=9)
    seq = ta.cmo(candles, period=9, sequential=True)

    assert round(single, 0) == -70
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_correl():
    candles = np.array(test_candles_19)

    single = ta.correl(candles)
    seq = ta.correl(candles, sequential=True)

    assert round(single, 2) == 0.58
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_correlation_cycle():
    candles = np.array(test_candles_19)

    single = ta.correlation_cycle(candles)
    assert type(single).__name__ == 'CC'
    assert round(single.real, 2) == 0.23
    assert round(single.imag, 2) == 0.38
    assert round(single.angle, 2) == -55.87
    assert round(single.state, 2) == -1

    seq = ta.correlation_cycle(candles, sequential=True)
    assert seq.real[-1] == single.real
    assert seq.imag[-1] == single.imag
    assert seq.angle[-1] == single.angle
    assert seq.state[-1] == single.state
    assert len(seq.real) == len(candles)
    assert len(seq.imag) == len(candles)
    assert len(seq.angle) == len(candles)
    assert len(seq.state) == len(candles)


def test_cvi():
    candles = np.array(test_candles_19)

    single = ta.cvi(candles)
    seq = ta.cvi(candles, sequential=True)

    assert round(single, 2) == 196.8
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_cwma():
    candles = np.array(test_candles_19)

    single = ta.cwma(candles)
    seq = ta.cwma(candles, sequential=True)

    assert round(single, 2) == 182.8
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_damiani_volatmeter():
    candles = np.array(test_candles_19)

    single = ta.damiani_volatmeter(candles)
    assert type(single).__name__ == 'DamianiVolatmeter'
    assert round(single.vol, 2) == 1.39
    assert round(single.anti, 2) == 0.93

    seq = ta.damiani_volatmeter(candles, sequential=True)
    assert seq.vol[-1] == single.vol
    assert seq.anti[-1] == single.anti
    assert len(seq.vol) == len(candles)
    assert len(seq.anti) == len(candles)


def test_dec_osc():
    candles = np.array(test_candles_19)
    single = ta.dec_osc(candles)
    seq = ta.dec_osc(candles, sequential=True)
    assert round(single, 0) == -20
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_decycler():
    candles = np.array(test_candles_19)
    single = ta.decycler(candles)
    seq = ta.decycler(candles, sequential=True)
    assert round(single, 0) == 233
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_dema():
    candles = np.array(test_candles_18)

    single = ta.dema(candles, 9)
    seq = ta.dema(candles, 9, sequential=True)

    assert round(single, 0) == 165
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_devstop():
    candles = np.array(test_candles_19)

    single = ta.devstop(candles)
    seq = ta.devstop(candles, sequential=True)

    assert round(single, 0) == 248.0
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_di():
    candles = np.array(test_candles_19)

    single = ta.di(candles, period=14)
    assert type(single).__name__ == 'DI'
    assert round(single.plus, 2) == 10.80
    assert round(single.minus, 1) == 45.3

    seq = ta.di(candles, period=14, sequential=True)
    assert seq.plus[-1] == single.plus
    assert seq.minus[-1] == single.minus
    assert len(seq.plus) == len(candles)
    assert len(seq.minus) == len(candles)


def test_dm():
    candles = np.array(test_candles_19)

    single = ta.dm(candles, period=14)
    assert type(single).__name__ == 'DM'
    assert round(single.plus, 2) == 36.78
    assert round(single.minus, 1) == 154.1

    seq = ta.dm(candles, period=14, sequential=True)
    assert seq.plus[-1] == single.plus
    assert seq.minus[-1] == single.minus
    assert len(seq.plus) == len(candles)
    assert len(seq.minus) == len(candles)


def test_donchian():
    candles = np.array(test_candles_19)

    single = ta.donchian(candles, period=20)
    seq = ta.donchian(candles, period=20, sequential=True)

    assert type(single).__name__ == 'DonchianChannel'
    assert round(single.upperband, 2) == 277.20
    assert round(single.middleband, 2) == 189.20
    assert round(single.lowerband, 2) == 101.20

    assert seq.middleband[-1] == single.middleband
    assert len(seq.upperband) == len(candles)
    assert len(seq.middleband) == len(candles)
    assert len(seq.lowerband) == len(candles)


def test_dpo():
    candles = np.array(test_candles_18)

    single = ta.dpo(candles)
    seq = ta.dpo(candles, sequential=True)

    assert round(single, 0) == 22
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_dti():
    candles = np.array(test_candles_19)

    single = ta.dti(candles)
    seq = ta.dti(candles, sequential=True)

    assert round(single, 2) == -32.6
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_dx():
    candles = np.array(test_candles_18)

    single = ta.dx(candles)
    seq = ta.dx(candles, sequential=True)

    assert round(single.adx, 2) == 83.12
    assert round(single.plusDI, 2) == 5.59
    assert round(single.minusDI, 2) == 58.93
    assert len(seq.adx) == len(candles)
    assert len(seq.plusDI) == len(candles)
    assert len(seq.minusDI) == len(candles)
    assert seq.adx[-1] == single.adx
    assert seq.plusDI[-1] == single.plusDI
    assert seq.minusDI[-1] == single.minusDI


def test_edcf():
    candles = np.array(test_candles_19)
    single = ta.edcf(candles)
    seq = ta.edcf(candles, sequential=True)

    assert round(single, 2) == 197.49
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_efi():
    candles = np.array(test_candles_19)
    single = ta.efi(candles)
    seq = ta.efi(candles, sequential=True)
    assert round(single, 0) == -51628073
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ema():
    close_prices = [
        204.23, 205.01, 196.9, 197.33, 198.7, 199.86, 202.23, 200.3, 212.3, 210.82603059, 220.84, 218.99,
        212.71, 211.01, 213.19, 212.99724894,
        212.67760477, 209.85, 187.2, 184.15, 176.99, 175.9, 178.99, 150.96, 133.85, 138.18, 126.32, 125.23,
        114.79,
        118.73, 110.74409879, 111.72, 124.04, 118.52, 113.64, 119.65, 117.11129288, 109.23, 110.77, 102.65,
        91.99
    ]
    candles = candles_from_close_prices(close_prices)

    single = ta.ema(candles, 8)
    seq = ta.ema(candles, 8, sequential=True)

    assert round(single, 3) == 108.546
    assert len(seq) == len(candles)
    assert seq[-1] == single
    assert np.isnan(ta.ema(candles, 400))


def test_emd():
    candles = np.array(test_candles_19)

    single = ta.emd(candles)
    seq = ta.emd(candles, sequential=True)

    assert type(single).__name__ == 'EMD'
    assert round(single.middleband, 2) == 3.12
    assert round(single.upperband, 2) == 1.21
    assert round(single.lowerband, 2) == -0.28

    assert seq.middleband[-1] == single.middleband
    assert seq.upperband[-1] == single.upperband
    assert seq.lowerband[-1] == single.lowerband
    assert len(seq.middleband) == len(candles)
    assert len(seq.upperband) == len(candles)
    assert len(seq.lowerband) == len(candles)


def test_emv():
    candles = np.array(test_candles_19)
    single = ta.emv(candles)
    seq = ta.emv(candles, sequential=True)
    assert round(single, 0) == -1
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_epma():
    candles = np.array(test_candles_19)

    single = ta.epma(candles)
    seq = ta.epma(candles, sequential=True)

    assert round(single, 2) == 175.31
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_er():
    candles = np.array(test_candles_19)
    single = ta.er(candles)
    seq = ta.er(candles, sequential=True)
    assert round(single, 2) == 0.02
    assert len(seq) == len(candles)
    assert round(seq[-1], 2) == round(single, 2)


def test_eri():
    candles = np.array(test_candles_19)
    single = ta.eri(candles)
    seq = ta.eri(candles, sequential=True)

    assert type(single).__name__ == 'ERI'
    assert round(single.bull, 2) == -7.14
    assert round(single.bear, 2) == -101.49

    assert seq.bull[-1] == single.bull
    assert len(seq.bull) == len(candles)
    assert len(seq.bear) == len(candles)


def test_fisher():
    candles = np.array(test_candles_19)
    single = ta.fisher(candles, period=9)
    seq = ta.fisher(candles, period=9, sequential=True)

    assert type(single).__name__ == 'FisherTransform'
    assert round(single.fisher, 2) == -1.77
    assert round(single.signal, 2) == -1.31

    assert seq.fisher[-1] == single.fisher
    assert len(seq.fisher) == len(candles)
    assert len(seq.signal) == len(candles)


def test_fosc():
    candles = np.array(test_candles_19)
    single = ta.fosc(candles)
    seq = ta.fosc(candles, sequential=True)
    assert round(single, 0) == -32
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_frama():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.frama(candles, window=10, SC=200, FC=10, )
    seq = ta.frama(candles, window=10, SC=200, FC=10, sequential=True)

    assert round(single, 0) == 219
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_fwma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.fwma(candles)
    seq = ta.fwma(candles, sequential=True)

    assert round(single, 0) == 161
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_gator():
    candles = np.array(test_candles_19)
    single = ta.gatorosc(candles)
    seq = ta.gatorosc(candles, sequential=True)

    assert type(single).__name__ == 'GATOR'
    assert round(single.upper, 2) == 2.39
    assert round(single.upper_change, 2) == 0.98
    assert round(single.lower, 2) == -13.44
    assert round(single.lower_change, 2) == 5.06

    assert seq.upper[-1] == single.upper
    assert len(seq.upper) == len(candles)


def test_gauss():
    candles = np.array(test_candles_19)
    single = ta.gauss(candles)
    seq = ta.gauss(candles, sequential=True)
    assert round(single, 0) == 190
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_heikin_ashi_candles():
    candles = np.array(test_candles_19)
    open_single, close_single, high_single, low_single = ta.heikin_ashi_candles(candles)
    open_seq, close_seq, high_seq, low_seq = ta.heikin_ashi_candles(candles, sequential=True)
    # SINGLES
    assert round(open_single, 2) == 197.68
    assert round(close_single, 2) == 149.8
    assert round(high_single, 2) == 197.68
    assert round(low_single, 2) == 101.2
    # SEQ
    assert len(open_seq) == len(candles)
    assert open_seq[-1] == open_single
    assert close_seq[-1] == close_single
    assert high_seq[-1] == high_single
    assert low_seq[-1] == low_single


def test_high_pass():
    candles = np.array(test_candles_19)
    single = ta.high_pass(candles)
    seq = ta.high_pass(candles, sequential=True)
    assert round(single, 0) == -106
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_high_pass_2_pole():
    candles = np.array(test_candles_19)
    single = ta.high_pass_2_pole(candles)
    seq = ta.high_pass_2_pole(candles, sequential=True)
    assert round(single, 0) == -101
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_hma():
    candles = np.array(test_candles_19)
    single = ta.hma(candles)
    seq = ta.hma(candles, sequential=True)

    assert round(single, 0) == 134
    assert len(seq) == len(candles)
    assert seq[-1] == single

def test_hurst():
    candles = np.array(test_candles_19)

    try:
        import numba
        no_numba = False
    except ImportError:
        no_numba = True

    if not no_numba:
        rs = ta.hurst_exponent(candles, method=0)
        assert round(rs, 2) == 0.51

    dma = ta.hurst_exponent(candles, method=1)
    dsod = ta.hurst_exponent(candles, method=2)

    assert round(dma, 2) == 0.26
    assert round(dsod, 2) == 0.5


def test_hwma():
    candles = np.array(test_candles_19)
    single = ta.hwma(candles)
    seq = ta.hwma(candles, sequential=True)

    assert round(single, 2) == 159.8
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ichimoku_cloud():
    candles = np.array(test_candles_15)

    ic = ta.ichimoku_cloud(candles)

    current_conversion_line, current_base_line, span_a, span_b = ic

    assert type(ic).__name__ == 'IchimokuCloud'

    assert (current_conversion_line, current_base_line, span_a,
            span_b) == (8861.59, 8861.59, 8466.385, 8217.45)


def test_ichimoku_cloud_seq():
    candles = np.array(test_candles_15)

    conversion_line, base_line, span_a, span_b, lagging_line, future_span_a, future_span_b = ta.ichimoku_cloud_seq(
        candles)
    seq = ta.ichimoku_cloud_seq(candles, sequential=True)

    assert type(seq).__name__ == 'IchimokuCloud'
    assert (conversion_line, base_line, span_a, span_b, lagging_line, future_span_a, future_span_b) == (
        seq.conversion_line[-1], seq.base_line[-1], seq.span_a[-1], seq.span_b[-1], seq.lagging_line[-1],
        seq.future_span_a[-1], seq.future_span_b[-1])
    assert (conversion_line, base_line, span_a, span_b, lagging_line, future_span_a, future_span_b) == (
        8861.59, 8861.59, 8465.25, 8204.715, 8730.0, 8861.59, 8579.49)
    assert len(seq.conversion_line) == len(candles)


def test_ift_rsi():
    # use the same candles as dema_candles
    candles = np.array(test_candles_19)

    single = ta.ift_rsi(candles)
    seq = ta.ift_rsi(candles, sequential=True)

    assert round(single, 2) == 0.89
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_itrend():
    candles = np.array(test_candles_19)
    single = ta.itrend(candles)
    seq = ta.itrend(candles, sequential=True)

    assert type(single).__name__ == 'ITREND'
    assert round(single.it, 0) == 223
    assert round(single.trigger, 0) == 182
    assert single.signal == -1

    assert seq.it[-1] == single.it
    assert seq.signal[-1] == single.signal
    assert seq.trigger[-1] == single.trigger
    assert len(seq.it) == len(candles)


def test_jma():
    # use the same candles as dema_candles
    candles = np.array(test_candles_19)

    single = ta.jma(candles)
    seq = ta.jma(candles, sequential=True)

    assert round(single, 2) == 156.72
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_jsa():
    # use the same candles as dema_candles
    candles = np.array(test_candles_19)

    single = ta.jsa(candles)
    seq = ta.jsa(candles, sequential=True)

    assert round(single, 2) == 172.26
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_kama():
    # use the same candles as dema_candles
    candles = np.array(test_candles_18)

    single = ta.kama(candles, 10)
    seq = ta.kama(candles, 10, sequential=True)

    assert round(single, 0) == 202
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_kaufmanstop():
    # use the same candles as dema_candles
    candles = np.array(test_candles_18)

    single = ta.kaufmanstop(candles)
    seq = ta.kaufmanstop(candles, sequential=True)

    assert round(single, 0) == 57
    assert len(seq) == len(candles)
    assert seq[-1] == single
    with pytest.raises(ValueError) as e:
        ta.kaufmanstop(candles, matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kaufmanstop indicator."

    with pytest.raises(ValueError) as e:
        ta.kaufmanstop(candles, matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kaufmanstop indicator."

    with pytest.raises(ValueError) as e:
        ta.kaufmanstop(candles, matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kaufmanstop indicator."

    with pytest.raises(ValueError) as e:
        ta.kaufmanstop(candles, matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kaufmanstop indicator."


def test_kdj():
    candles = np.array(test_candles_19)

    kd = ta.kdj(candles)
    k, d, j = kd
    assert type(kd).__name__ == 'KDJ'
    assert round(k, 1) == 13.3
    assert round(d, 1) == 15.7
    assert round(j, 1) == 8.6

    seq_kd = ta.kdj(candles, sequential=True)
    assert seq_kd.k[-1] == k
    assert len(seq_kd.k) == len(candles)
    assert len(seq_kd.d) == len(candles)
    assert len(seq_kd.j) == len(candles)
    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowk_matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowk_matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowd_matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowd_matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowk_matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowk_matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowd_matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."

    with pytest.raises(ValueError) as e:
        ta.kdj(candles, slowd_matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in kdj indicator."


def test_kelner_channels():
    candles = np.array(test_candles_13)

    kc = ta.keltner(candles)
    u, m, l = kc
    assert type(kc).__name__ == 'KeltnerChannel'
    assert round(u, 1) == 145.0
    assert round(m, 1) == 139.7
    assert round(l, 1) == 134.4

    seq_kc = ta.keltner(candles, sequential=True)
    assert seq_kc.upperband[-1] == u
    assert len(seq_kc.upperband) == len(candles)
    assert len(seq_kc.middleband) == len(candles)
    assert len(seq_kc.lowerband) == len(candles)


def test_kst():
    candles = np.array(test_candles_19)

    single = ta.kst(candles)
    seq = ta.kst(candles, sequential=True)

    assert type(single).__name__ == 'KST'
    assert round(single.line, 2) == -93.38
    assert round(single.signal, 2) == 31.1

    assert seq.line[-1] == single.line
    assert seq.signal[-1] == single.signal
    assert len(seq.line) == len(candles)
    assert len(seq.signal) == len(candles)


def test_kurtosis():
    candles = np.array(test_candles_19)

    single = ta.kurtosis(candles)
    seq = ta.kurtosis(candles, sequential=True)

    assert round(single, 2) == -0.22
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_kvo():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.kvo(candles)
    seq = ta.kvo(candles, sequential=True)

    assert round(single / 10000000, 2) == -0.88
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_linearreg():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.linearreg(candles)
    seq = ta.linearreg(candles, sequential=True)

    assert round(single, 2) == 179.56
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_linearreg_angle():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.linearreg_angle(candles)
    seq = ta.linearreg_angle(candles, sequential=True)

    assert round(single, 2) == -78.42
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_linearreg_intercept():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.linearreg_intercept(candles)
    seq = ta.linearreg_intercept(candles, sequential=True)

    assert round(single, 2) == 242.98
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_linearreg_slope():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.linearreg_slope(candles)
    seq = ta.linearreg_slope(candles, sequential=True)

    assert round(single, 2) == -4.88
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_lrsi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.lrsi(candles)
    seq = ta.lrsi(candles, sequential=True)

    assert round(single, 2) == 0.1
    assert round(seq[-2], 2) == 0.04
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)
    single = ta.ma(candles, matype=9)
    seq = ta.ma(candles, matype=9, sequential=True)

    seq_average = ta.ma(seq, matype=9, sequential=True)

    assert round(single, 2) == 166.99
    assert round(seq[-2], 2) == 203.56
    assert round(seq_average[-2], 2) == 212.12
    assert len(seq) == len(candles)
    assert len(seq_average) == len(candles)
    assert seq[-1] == single


def test_maaq():
    candles = np.array(test_candles_19)

    single = ta.maaq(candles)
    seq = ta.maaq(candles, sequential=True)

    assert round(single, 2) == 205.95
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mab():
    candles = np.array(test_candles_11)

    bb = ta.mab(candles)
    u, m, l = bb
    assert type(bb).__name__ == 'MAB'
    assert round(u, 1) == 142.8
    assert round(m, 1) == 140.0
    assert round(l, 1) == 130.1

    seq = ta.mab(candles, sequential=True)
    assert seq.upperband[-1] == u
    assert len(seq.upperband) == len(candles)
    assert len(seq.middleband) == len(candles)
    assert len(seq.lowerband) == len(candles)


def test_macd():
    candles = np.array(test_candles_19)

    single = ta.macd(candles, fast_period=12, slow_period=26, signal_period=9)
    seq = ta.macd(candles, fast_period=12, slow_period=26, signal_period=9, sequential=True)

    assert type(single).__name__ == 'MACD'
    assert round(single.macd, 2) == -15.32
    assert round(single.signal, 2) == -4.10
    assert round(single.hist, 2) == -11.22

    assert seq.macd[-1] == single.macd
    assert len(seq.macd) == len(candles)
    assert len(seq.signal) == len(candles)
    assert len(seq.hist) == len(candles)

def test_mama():
    candles = np.array(test_candles_19)

    mama = ta.mama(candles, 0.5, 0.05)
    assert type(mama).__name__ == 'MAMA'
    assert round(mama.mama, 2) == 206.78
    assert round(mama.fama, 2) == 230.26

    seq_mama = ta.mama(candles, 0.5, 0.05, sequential=True)
    assert seq_mama.mama[-1] == mama.mama
    assert len(seq_mama.mama) == len(candles)
    assert len(seq_mama.fama) == len(candles)


def test_marketfi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.marketfi(candles)
    seq = ta.marketfi(candles, sequential=True)

    assert round(single * 100000, 2) == 2.47
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mass():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.mass(candles)
    seq = ta.mass(candles, sequential=True)

    assert round(single, 2) == 5.76
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mcginley_dynamic():
    candles = np.array(test_candles_19)

    single = ta.mcginley_dynamic(candles)
    seq = ta.mcginley_dynamic(candles, sequential=True)
    assert round(single, 2) == 107.82
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mean_ad():
    candles = np.array(test_candles_19)

    single = ta.mean_ad(candles)
    seq = ta.mean_ad(candles, sequential=True)

    assert round(single, 2) == 23.82
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_median_ad():
    candles = np.array(test_candles_19)

    single = ta.median_ad(candles)
    seq = ta.median_ad(candles, sequential=True)

    assert round(single, 2) == 6.86
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_medprice():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.medprice(candles)
    seq = ta.medprice(candles, sequential=True)

    assert round(single, 1) == 148.4
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mfi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.mfi(candles, period=9)
    seq = ta.mfi(candles, period=9, sequential=True)

    assert round(single, 1) == 31.2
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_midpoint():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.midpoint(candles)
    seq = ta.midpoint(candles, sequential=True)

    assert round(single, 1) == 176.4
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_midprice():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.midprice(candles)
    seq = ta.midprice(candles, sequential=True)

    assert round(single, 1) == 176.6
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_minmax():
    candles = np.array(test_candles_19)
    single = ta.minmax(candles)
    seq = ta.minmax(candles, sequential=True)

    assert type(single).__name__ == 'EXTREMA'
    assert round(seq.is_max[-6], 2) == 251.93
    assert round(seq.is_min[-15], 2) == 210
    assert round(single.last_max, 2) == 251.93
    assert round(single.last_min, 2) == 210

    assert seq.last_max[-1] == single.last_max
    assert seq.last_min[-1] == single.last_min
    assert len(seq.is_min) == len(candles)


def test_mom():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.mom(candles, period=9)
    seq = ta.mom(candles, period=9, sequential=True)

    assert round(single, 2) == -116.09
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_mwdx():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.mwdx(candles)
    seq = ta.mwdx(candles, sequential=True)

    assert round(single, 2) == 192.11
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_natr():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.natr(candles, period=14)
    seq = ta.natr(candles, period=14, sequential=True)

    assert round(single, 2) == 22.55
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_nma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.nma(candles)
    seq = ta.nma(candles, sequential=True)

    assert round(single, 2) == 185.39
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_nvi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.nvi(candles)
    seq = ta.nvi(candles, sequential=True)

    assert round(single, 2) == 722.58
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_obv():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.obv(candles)
    seq = ta.obv(candles, sequential=True)

    assert round(single / 1000000, 0) == -6
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_pfe():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.pfe(candles)
    seq = ta.pfe(candles, sequential=True)

    assert round(single, 2) == -211.85
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_pivot():
    candles = np.array(test_candles_19)
    single = ta.pivot(candles, mode=0)
    seq = ta.pivot(candles, mode=0, sequential=True)

    assert type(single).__name__ == 'PIVOT'

    assert seq.r1[-1] == single.r1
    assert len(seq.r1) == len(candles)
    assert len(seq.r2) == len(candles)
    assert len(seq.r3) == len(candles)
    assert len(seq.r4) == len(candles)
    assert len(seq.pp) == len(candles)
    assert len(seq.s1) == len(candles)
    assert len(seq.s2) == len(candles)
    assert len(seq.s3) == len(candles)
    assert len(seq.s4) == len(candles)


def test_pivot1():
    candles = np.array(test_candles_19)
    single = ta.pivot(candles, mode=1)
    seq = ta.pivot(candles, mode=1, sequential=True)

    assert type(single).__name__ == 'PIVOT'

    assert seq.r1[-1] == single.r1
    assert len(seq.r1) == len(candles)
    assert len(seq.r2) == len(candles)
    assert len(seq.r3) == len(candles)
    assert len(seq.r4) == len(candles)
    assert len(seq.pp) == len(candles)
    assert len(seq.s1) == len(candles)
    assert len(seq.s2) == len(candles)
    assert len(seq.s3) == len(candles)
    assert len(seq.s4) == len(candles)


def test_pivot2():
    candles = np.array(test_candles_19)
    single = ta.pivot(candles, mode=2)
    seq = ta.pivot(candles, mode=2, sequential=True)

    assert type(single).__name__ == 'PIVOT'

    assert seq.r1[-1] == single.r1
    assert len(seq.r1) == len(candles)
    assert len(seq.r2) == len(candles)
    assert len(seq.r3) == len(candles)
    assert len(seq.r4) == len(candles)
    assert len(seq.pp) == len(candles)
    assert len(seq.s1) == len(candles)
    assert len(seq.s2) == len(candles)
    assert len(seq.s3) == len(candles)
    assert len(seq.s4) == len(candles)


def test_pivot3():
    candles = np.array(test_candles_19)
    single = ta.pivot(candles, mode=3)
    seq = ta.pivot(candles, mode=3, sequential=True)

    assert type(single).__name__ == 'PIVOT'

    assert seq.r1[-1] == single.r1
    assert len(seq.r1) == len(candles)
    assert len(seq.r2) == len(candles)
    assert len(seq.r3) == len(candles)
    assert len(seq.r4) == len(candles)
    assert len(seq.pp) == len(candles)
    assert len(seq.s1) == len(candles)
    assert len(seq.s2) == len(candles)
    assert len(seq.s3) == len(candles)
    assert len(seq.s4) == len(candles)


def test_pivot4():
    candles = np.array(test_candles_19)
    single = ta.pivot(candles, mode=4)
    seq = ta.pivot(candles, mode=4, sequential=True)

    assert type(single).__name__ == 'PIVOT'

    assert seq.r1[-1] == single.r1
    assert len(seq.r1) == len(candles)
    assert len(seq.r2) == len(candles)
    assert len(seq.r3) == len(candles)
    assert len(seq.r4) == len(candles)
    assert len(seq.pp) == len(candles)
    assert len(seq.s1) == len(candles)
    assert len(seq.s2) == len(candles)
    assert len(seq.s3) == len(candles)
    assert len(seq.s4) == len(candles)


def test_pma():
    candles = np.array(test_candles_19)

    single = ta.pma(candles)
    assert type(single).__name__ == 'PMA'
    assert round(single.predict, 2) == 171.05
    assert round(single.trigger, 2) == 190.91

    seq = ta.pma(candles, sequential=True)
    assert seq.predict[-1] == single.predict
    assert seq.trigger[-1] == single.trigger
    assert len(seq.predict) == len(candles)
    assert len(seq.trigger) == len(candles)


def test_ppo():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.ppo(candles, fast_period=12, slow_period=26, matype=1)
    seq = ta.ppo(candles, fast_period=12, slow_period=26, matype=1, sequential=True)

    assert round(single, 0) == -7
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_pvi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.pvi(candles)
    seq = ta.pvi(candles, sequential=True)

    assert round(single, 0) == 661
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_pwma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.pwma(candles)
    seq = ta.pwma(candles, sequential=True)

    assert round(single, 2) == 193.82
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_qstick():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.qstick(candles)
    seq = ta.qstick(candles, sequential=True)

    assert round(single, 0) == -26.0
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_reflex():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.reflex(candles)
    seq = ta.reflex(candles, sequential=True)

    assert round(single, 2) == -0.55
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rma():
    # use the same candles as wavetrend, Uses 'high' series instead of close
    candles = np.array(wavetrend_candles)

    single = ta.rma(candles)
    seq = ta.rma(candles, sequential=True)

    assert round(single, 2) == 31887.21
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_roc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.roc(candles, period=14)
    seq = ta.roc(candles, period=14, sequential=True)

    assert round(single, 2) == -52.67
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rocp():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.rocp(candles, period=14)
    seq = ta.rocp(candles, period=14, sequential=True)

    assert round(single, 2) == -0.53
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rocr():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.rocr(candles, period=14)
    seq = ta.rocr(candles, period=14, sequential=True)

    assert round(single, 2) == 0.47
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rocr100():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.rocr100(candles, period=14)
    seq = ta.rocr100(candles, period=14, sequential=True)

    assert round(single, 2) == 47.33
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_roofing():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.roofing(candles)
    seq = ta.roofing(candles, sequential=True)

    assert round(single, 0) == -36
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rsi():
    candles = np.array(test_candles_14)

    single = ta.rsi(candles)
    seq = ta.rsi(candles, sequential=True)

    assert round(single, 2) == 57.84
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rsmk():
    candles = np.array(test_candles_4)
    candles2 = np.array(test_candles_19)

    rsmk = ta.rsmk(candles, candles2)
    assert type(rsmk).__name__ == 'RSMK'
    assert round(rsmk.indicator, 2) == 2.1
    assert round(rsmk.signal, 2) == -31.56

    rsmk_seq = ta.rsmk(candles, candles2, sequential=True)
    assert rsmk_seq.indicator[-1] == rsmk.indicator
    assert rsmk_seq.signal[-1] == rsmk.signal
    assert len(rsmk_seq.indicator) == len(candles)
    assert len(rsmk_seq.signal) == len(candles)


def test_rsx():
    candles = np.array(test_candles_19)

    single = ta.rsx(candles)
    seq = ta.rsx(candles, sequential=True)

    assert round(single, 2) == 27.81
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_rvi():
    candles = np.array(test_candles_19)

    single = ta.rvi(candles)
    seq = ta.rvi(candles, sequential=True)

    assert round(single, 2) == 27.99
    assert len(seq) == len(candles)
    assert seq[-1] == single
    with pytest.raises(ValueError) as e:
        ta.rvi(candles, matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in rvi indicator."

    with pytest.raises(ValueError) as e:
        ta.rvi(candles, matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in rvi indicator."

    with pytest.raises(ValueError) as e:
        ta.rvi(candles, matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in rvi indicator."

    with pytest.raises(ValueError) as e:
        ta.rvi(candles, matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in rvi indicator."


def test_safezonestop():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.safezonestop(candles)
    seq = ta.safezonestop(candles, sequential=True)

    assert round(single, 2) == -39.15
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_sar():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.sar(candles, acceleration=0.02, maximum=0.2)
    seq = ta.sar(candles, acceleration=0.02, maximum=0.2, sequential=True)

    assert round(single, 2) == 243.15
    assert len(seq) == len(candles)
    assert seq[-1] == single



def test_sinwma():
    candles = np.array(test_candles_19)

    single = ta.sinwma(candles)
    seq = ta.sinwma(candles, sequential=True)

    assert round(single, 2) == 218.86
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_skew():
    candles = np.array(test_candles_19)

    single = ta.skew(candles)
    seq = ta.skew(candles, sequential=True)

    assert round(single, 2) == -1.05
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_sma():
    close_prices = [22.27, 22.19, 22.08, 22.17, 22.18, 22.13, 22.23, 22.43, 22.24, 22.29]
    candles = candles_from_close_prices(close_prices)

    single = ta.sma(candles, 10)
    seq = ta.sma(candles, 10, sequential=True)

    assert round(single, 2) == 22.22
    assert len(seq) == len(candles)
    assert seq[-1] == single
    assert np.isnan(ta.sma(candles, 30))


def test_smma():
    candles = np.array(test_candles_19)
    single = ta.smma(candles)
    seq = ta.smma(candles, sequential=True)

    assert round(single, 0) == 192
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_sqwma():
    candles = np.array(test_candles_19)
    single = ta.sqwma(candles)
    seq = ta.sqwma(candles, sequential=True)

    assert round(single, 2) == 191.02
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_srsi():
    candles = np.array(test_candles_4)
    period = 14

    srsi = ta.srsi(candles)
    k, d = srsi
    assert type(srsi).__name__ == 'StochasticRSI'
    assert round(k, 2) == 21.36
    assert round(d, 2) == 12.4

    srsi = ta.srsi(candles, period=period, sequential=True)
    assert srsi.d[-1] == d
    assert srsi.k[-1] == k
    assert len(srsi.d) == len(candles)
    assert len(srsi.k) == len(candles)


def test_srwma():
    candles = np.array(test_candles_19)
    single = ta.srwma(candles)
    seq = ta.srwma(candles, sequential=True)

    assert round(single, 2) == 205.38
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_stc():
    candles = np.array(test_candles_19)
    single = ta.stc(candles)
    seq = ta.stc(candles, sequential=True)

    assert round(single, 2) == 0.0
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_stddev():
    candles = np.array(test_candles_19)
    single = ta.stddev(candles)
    seq = ta.stddev(candles, sequential=True)

    assert round(single, 0) == 37
    assert len(seq) == len(candles)
    assert seq[-1] == single

    closes = candles[:, 2]
    single = ta.stddev(closes)
    seq = ta.stddev(closes, sequential=True)

    assert round(single, 0) == 37
    assert seq[-1] == single


def test_stoch():
    candles = np.array(test_candles_3)

    stoch = ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0)
    k, d = stoch
    assert type(stoch).__name__ == 'Stochastic'
    assert round(k, 2) == 53.68
    assert round(d, 2) == 49.08

    stoch = ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0,
                     sequential=True)
    assert stoch.d[-1] == d
    assert stoch.k[-1] == k
    assert len(stoch.d) == len(candles)
    assert len(stoch.k) == len(candles)
    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=24, slowd_period=3, slowd_matype=0)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=29, slowd_period=3, slowd_matype=0)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=24, slowd_period=3, slowd_matype=0, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=29, slowd_period=3, slowd_matype=0, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."

    with pytest.raises(ValueError) as e:
        ta.stoch(candles, fastk_period=14, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator."


def test_stochf():
    candles = np.array(test_candles_19)

    single = ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=0)
    seq = ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=0, sequential=True)

    assert type(single).__name__ == 'StochasticFast'
    assert round(single.k, 2) == 4.87
    assert round(single.d, 2) == 13.5

    assert seq.k[-1] == single.k
    assert len(seq.k) == len(candles)
    assert len(seq.d) == len(candles)
    
    with pytest.raises(ValueError) as e:
        ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=24)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochf indicator."

    with pytest.raises(ValueError) as e:
        ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=29)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochf indicator."

    with pytest.raises(ValueError) as e:
        ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=24, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochf indicator."

    with pytest.raises(ValueError) as e:
        ta.stochf(candles, fastk_period=5, fastd_period=3, fastd_matype=29, sequential=True)
    assert str(e.value) == "VWMA (matype 24) and VWAP (matype 29) cannot be used in stochf indicator."
    


def test_supersmoother():
    candles = np.array(test_candles_19)
    single = ta.supersmoother(candles)
    seq = ta.supersmoother(candles, sequential=True)
    assert round(single, 0) == 201
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_supersmoother_3_pole():
    candles = np.array(test_candles_19)
    single = ta.supersmoother_3_pole(candles)
    seq = ta.supersmoother_3_pole(candles, sequential=True)
    assert round(single, 0) == 207
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_supertrend():
    candles = np.array(test_candles_19)

    single = ta.supertrend(candles, period=10, factor=3)
    seq = ta.supertrend(candles, period=10, factor=3, sequential=True)

    assert type(single).__name__ == 'SuperTrend'
    assert round(single.trend, 2) == 228.45
    assert seq.changed[-16] == True
    assert seq.changed[-1] == False

    assert seq.trend[-1] == single.trend
    assert len(seq.trend) == len(candles)
    assert len(seq.changed) == len(candles)


def test_swma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.swma(candles)
    seq = ta.swma(candles, sequential=True)

    assert round(single, 2) == 189.35
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_t3():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.t3(candles, period=5, vfactor=0.7)
    seq = ta.t3(candles, period=5, vfactor=0.7, sequential=True)

    assert round(single, 0) == 194
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_tema():
    # use the same candles as trix_candles
    candles = np.array(test_candles_17)

    single = ta.tema(candles)
    seq = ta.tema(candles, sequential=True)

    assert round(single, 2) == 213.2
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_trange():
    candles = np.array(test_candles_19)

    single = ta.trange(candles)
    seq = ta.trange(candles, sequential=True)

    assert round(single, 2) == 94.35
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_trendflex():
    candles = np.array(test_candles_19)

    single = ta.trendflex(candles)
    seq = ta.trendflex(candles, sequential=True)

    assert round(single, 2) == -1.48
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_trima():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.trima(candles, period=9)
    seq = ta.trima(candles, period=9, sequential=True)

    assert round(single, 0) == 211
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_trix():
    candles = np.array(test_candles_17)

    single = ta.trix(candles)
    seq = ta.trix(candles, sequential=True)

    assert round(single, 2) == 30.28
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_tsf():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.tsf(candles)
    seq = ta.tsf(candles, sequential=True)

    assert round(single, 1) == 174.7
    assert len(seq) == len(candles)
    assert round(seq[-1], 2) == round(single, 2)


def test_tsi():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.tsi(candles)
    seq = ta.tsi(candles, sequential=True)

    assert round(single, 1) == -20.5
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ttm_trend():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.ttm_trend(candles)
    seq = ta.ttm_trend(candles, sequential=True)

    assert single == False
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_typprice():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.typprice(candles)
    seq = ta.typprice(candles, sequential=True)

    assert round(single, 1) == 134.9
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ui():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.ui(candles)
    seq = ta.ui(candles, sequential=True)

    assert round(single, 1) == 23.7
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_ultosc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.ultosc(candles, timeperiod1=7, timeperiod2=14, timeperiod3=28)
    seq = ta.ultosc(candles, timeperiod1=7, timeperiod2=14, timeperiod3=28, sequential=True)

    assert round(single, 2) == 31.37
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_var():
    candles = np.array(test_candles_16)
    single = ta.var(candles)
    seq = ta.var(candles, sequential=True)

    assert round(single, 2) == 69.96
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vi():
    candles = np.array(test_candles_19)

    single = ta.vi(candles)
    seq = ta.vi(candles, sequential=True)

    assert type(single).__name__ == 'VI'
    assert round(single.plus, 2) == 0.66
    assert round(single.minus, 2) == 1.13

    assert seq.plus[-1] == single.plus
    assert len(seq.plus) == len(candles)
    assert len(seq.minus) == len(candles)


def test_vidya():
    candles = np.array(test_candles_16)
    single = ta.vidya(candles)
    seq = ta.vidya(candles, sequential=True)

    assert round(single, 2) == 200.84
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vlma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.vlma(candles)
    seq = ta.vlma(candles, sequential=True)

    assert round(single, 2) == 208.1
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vosc():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.vosc(candles)
    seq = ta.vosc(candles, sequential=True)

    assert round(single, 2) == 38.18
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_voss():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.voss(candles)
    seq = ta.voss(candles, sequential=True)

    assert type(single).__name__ == 'VossFilter'
    assert round(single.voss, 2) == -30.71
    assert round(single.filt, 2) == -5.98

    assert seq.voss[-1] == single.voss
    assert seq.filt[-1] == single.filt
    assert len(seq.voss) == len(candles)
    assert len(seq.filt) == len(candles)


def test_vpci():
    candles = np.array(test_candles_19)
    single = ta.vpci(candles)
    seq = ta.vpci(candles, sequential=True)

    assert round(single.vpci, 2) == -29.46
    assert round(single.vpcis, 2) == -14.4
    assert len(seq.vpci) == len(candles)
    assert seq.vpci[-1] == single.vpci


def test_vpt():
    candles = np.array(test_candles_19)
    single = ta.vpt(candles)
    seq = ta.vpt(candles, sequential=True)

    assert round(single, 2) == -1733928.99
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vpwma():
    candles = np.array(test_candles_19)
    single = ta.vpwma(candles)
    seq = ta.vpwma(candles, sequential=True)

    assert round(single, 2) == 206.52
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vwap():
    candles = np.array(test_candles_19)
    single = ta.vwap(candles)
    seq = ta.vwap(candles, sequential=True)

    assert round(single, 2) == 134.86
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vwma():
    candles = np.array(test_candles_16)
    single = ta.vwma(candles)
    seq = ta.vwma(candles, sequential=True)

    assert round(single, 2) == 195.86
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_vwmacd():
    candles = np.array(test_candles_19)

    single = ta.vwmacd(candles, fast_period=12, slow_period=26, signal_period=9)
    seq = ta.vwmacd(candles, fast_period=12, slow_period=26, signal_period=9, sequential=True)

    assert type(single).__name__ == 'VWMACD'
    assert round(single.macd, 2) == -31.37
    assert round(single.signal, 2) == -20.72
    assert round(single.hist, 2) == -10.65

    assert seq.macd[-1] == single.macd
    assert len(seq.macd) == len(candles)
    assert len(seq.signal) == len(candles)
    assert len(seq.hist) == len(candles)


def test_wad():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.wad(candles)
    seq = ta.wad(candles, sequential=True)

    assert round(single, 2) == -122.14
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_wclprice():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.wclprice(candles)
    seq = ta.wclprice(candles, sequential=True)

    assert round(single, 2) == 128.1
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_wilders():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.wilders(candles)
    seq = ta.wilders(candles, sequential=True)

    assert round(single, 2) == 192.11
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_willr():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.willr(candles, period=9)
    seq = ta.willr(candles, period=9, sequential=True)

    assert round(single, 2) == -95.61
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_wma():
    # use the same candles as mama_candles
    candles = np.array(test_candles_19)

    single = ta.wma(candles, period=9)
    seq = ta.wma(candles, period=9, sequential=True)

    assert round(single, 2) == 189.13
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_wt():
    candles = np.array(wavetrend_candles)
    single = ta.wt(candles)
    seq = ta.wt(candles, sequential=True)

    assert type(single).__name__ == 'Wavetrend'
    assert round(single.wt1, 2) == -27.25
    assert round(single.wt2, 2) == -15.51
    assert not single.wtCrossUp
    assert single.wtCrossDown
    assert single.wtCrossUp != single.wtCrossDown
    assert not single.wtOversold
    assert not single.wtOverbought
    assert round(single.wtVwap, 2) == -11.74

    assert seq.wt1[-1] == single.wt1
    assert seq.wt2[-1] == single.wt2
    assert seq.wtCrossUp[-1] is single.wtCrossUp
    assert seq.wtCrossDown[-1] is single.wtCrossDown
    assert seq.wtOversold[-1] is single.wtOversold
    assert seq.wtOverbought[-1] is single.wtOverbought
    assert seq.wtVwap[-1] == single.wtVwap
    assert len(seq.wt1) == len(candles)


def test_zlema():
    candles = np.array(test_candles_19)
    single = ta.zlema(candles)
    seq = ta.zlema(candles, sequential=True)

    assert round(single, 0) == 189
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_zscore():
    candles = np.array(test_candles_19)
    single = ta.zscore(candles)
    seq = ta.zscore(candles, sequential=True)

    assert round(single, 1) == -3.2
    assert len(seq) == len(candles)
    assert seq[-1] == single


def test_waddah_attr_explosion():
    candles = np.array(test_candles_19)
    single = ta.waddah_attar_explosion(candles)

    assert round(single[0]) == 135
    assert round(single[1]) == -827
    assert round(single[2]) == -1


def test_stiffness():
    candles = np.array(test_candles_19)

    assert round(ta.stiffness(candles)) == 96


def test_ttm_squeeze():
    candles = np.array(test_candles_19)
    result = ta.ttm_squeeze(candles)

    assert result == True


def test_support_resistance_with_breaks():
    candles = np.array(test_candles_19)
    result = ta.support_resistance_with_breaks(candles)

    assert result.support == 116.26
    assert result.resistance == 288.41
    assert result.red_break == False
    assert result.green_break == False
    assert result.bear_wick == False
    assert result.bull_wick == False


def test_squeeze_momentum():
    candles = np.array(test_candles_19)
    result = ta.squeeze_momentum(candles, sequential=False)

    assert result.squeeze == 1
    assert round(result.momentum, 2) == -52.23
    assert result.momentum_signal == -1

    seq_result = ta.squeeze_momentum(candles)

    assert seq_result.squeeze[0] == 0
    assert seq_result.squeeze[-2] == 1
    assert seq_result.momentum_signal[0] == -2
    assert seq_result.momentum_signal[-2] == -1
    assert round(seq_result.momentum[-10], 2) == -37.97
    assert round(seq_result.momentum[-2], 2) == -40.98


def test_hull_suit():
    candles = np.array(test_candles_19)
    result = ta.hull_suit(candles, sequential=False)

    assert round(result.s_hull, 1) == 243.7
    assert round(result.m_hull, 1) == 230.6
    assert result.signal == 'sell'


def test_volume():
    candles = np.array(fake_test_candles, dtype=np.float64)
    single = ta.volume(candles, period=5)
    seq = ta.volume(candles, period=5, sequential=True)

    assert type(single).__name__ == 'Volume'
    assert round(single.volume, 0) == 22
    assert round(single.ma, 0) == 20

    assert seq.volume[-1] == single.volume
    assert seq.ma[-1] == single.ma
    assert len(seq.volume) == len(candles)
    assert len(seq.ma) == len(candles)



from typing import Union

import numpy as np
from jesse.helpers import slice_candles
from jesse_rust import adosc as adosc_rust


def adosc(candles: np.ndarray, fast_period: int = 3, slow_period: int = 10, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    ADOSC - Chaikin A/D Oscillator (Rust accelerated version)

    :param candles: np.ndarray of candles
    :param fast_period: int - default: 3
    :param slow_period: int - default: 10
    :param sequential: bool - default: False
    :return: float or np.ndarray
    """
    candles = slice_candles(candles, sequential)
    
    # Convert to float64 for Rust compatibility
    candles_f64 = np.asarray(candles, dtype=np.float64)
    
    # Call the Rust implementation
    result = adosc_rust(candles_f64, fast_period, slow_period)

    return result if sequential else result[-1]



from typing import Union

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from scipy import stats

from jesse.helpers import get_candle_source, same_length, slice_candles


def median_ad(candles: np.ndarray, period: int = 5, source_type: str = "hl2", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Median Absolute Deviation

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "hl2"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    if len(candles.shape) == 1:
      source = candles
    else:
      candles = slice_candles(candles, sequential)
      source = get_candle_source(candles, source_type=source_type)

    swv = sliding_window_view(source, window_shape=period)
    median_abs_deviation = stats.median_abs_deviation(swv, axis=-1)
    res = same_length(source, median_abs_deviation)

    return res if sequential else res[-1]



from typing import Optional
from fastapi import APIRouter, Header, BackgroundTasks
from fastapi.responses import JSONResponse

from jesse.services import auth as authenticator
from jesse.services.multiprocessing import process_manager
from jesse.services.web import LoginRequestJson
import jesse.helpers as jh

router = APIRouter(prefix="/auth", tags=["Authentication"])


@router.post("/login")
def login(json_request: LoginRequestJson):
    """
    Authenticate user with password and return a token
    """
    return authenticator.password_to_token(json_request.password)


@router.post("/terminate-all")
async def terminate_all(authorization: Optional[str] = Header(None)):
    """
    Terminate all running processes
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.flush()
    return JSONResponse({'message': 'terminating all tasks...'})


@router.post("/shutdown")
async def shutdown(background_tasks: BackgroundTasks, authorization: Optional[str] = Header(None)):
    """
    Shutdown the application
    """
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    background_tasks.add_task(jh.terminate_app)
    return JSONResponse({'message': 'Shutting down...'})



import peewee
from jesse.services.db import database

if database.is_closed():
    database.open_connection()


class NotificationApiKeys(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    name = peewee.CharField(unique=True)
    driver = peewee.CharField()  # notification driver (Telegram, Discord, Slack)
    fields = peewee.TextField()  # for storing the fields as a JSON string
    created_at = peewee.DateTimeField()

    class Meta:
        from jesse.services.db import database

        database = database.db

    def __init__(self, attributes=None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])


# if database is open, create the table
if database.is_open():
    NotificationApiKeys.create_table()



import aioredis
import redis as sync_redis_lib
import simplejson as json
import asyncio
import jesse.helpers as jh
from jesse.libs.custom_json import NpEncoder
import os
import base64
from jesse.services.env import ENV_VALUES


async def init_redis():
    return await aioredis.create_redis_pool(
        address=(ENV_VALUES['REDIS_HOST'], ENV_VALUES['REDIS_PORT']),
        password=ENV_VALUES['REDIS_PASSWORD'] or None,
        db=int(ENV_VALUES.get('REDIS_DB') or 0),
    )


async_redis = None
sync_redis = None
if jh.is_jesse_project():
    if not jh.is_notebook():
        async_redis = asyncio.run(init_redis())
        sync_redis = sync_redis_lib.Redis(
            host=ENV_VALUES['REDIS_HOST'], port=ENV_VALUES['REDIS_PORT'], db=int(ENV_VALUES.get('REDIS_DB') or 0),
            password=ENV_VALUES['REDIS_PASSWORD'] if ENV_VALUES['REDIS_PASSWORD'] else None
        )


def sync_publish(event: str, msg, compression: bool = False):
    if jh.is_unit_testing():
        raise EnvironmentError('sync_publish() should be NOT called during testing. There must be something wrong')

    if compression:
        msg = jh.gzip_compress(msg)
        # Encode the compressed message using Base64
        msg = base64.b64encode(msg).decode('utf-8')

    sync_redis.publish(
        f"{ENV_VALUES['APP_PORT']}:channel:1", json.dumps({
            'id': os.getpid(),
            'event': f'{jh.app_mode()}.{event}',
            'is_compressed': compression,
            'data': msg
        }, ignore_nan=True, cls=NpEncoder)
    )


async def async_publish(event: str, msg, compression: bool = False):
    if jh.is_unit_testing():
        raise EnvironmentError('sync_publish() should be NOT called during testing. There must be something wrong')

    if compression:
        msg = jh.gzip_compress(msg)
        # Encode the compressed message using Base64
        msg = base64.b64encode(msg).decode('utf-8')

    await async_redis.publish(
        f"{ENV_VALUES['APP_PORT']}:channel:1", json.dumps({
            'id': os.getpid(),
            'event': f'{jh.app_mode()}.{event}',
            'is_compressed': compression,
            'data': msg
        }, ignore_nan=True, cls=NpEncoder)
    )


def is_process_active(client_id: str) -> bool:
    if jh.is_unit_testing():
        return False

    return sync_redis.sismember(f"{ENV_VALUES['APP_PORT']}|active-processes", client_id)



from typing import Union

import numpy as np
from jesse.helpers import slice_candles
from numpy.lib.stride_tricks import sliding_window_view


def midprice(candles: np.ndarray, period: int = 14, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    MIDPRICE - Midpoint Price over period

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    high = candles[:, 3]
    low = candles[:, 4]

    if sequential:
        n = len(candles)
        if n < period:
            return np.full(n, np.nan)

        # Create sliding windows for high and low prices
        windows_high = sliding_window_view(high, window_shape=period)
        windows_low = sliding_window_view(low, window_shape=period)

        # Calculate midprice for each window
        midprices = (np.max(windows_high, axis=1) + np.min(windows_low, axis=1)) / 2

        # Prepend NaN for the initial period-1 values to match the typical TA-Lib output length
        result = np.concatenate((np.full(period - 1, np.nan), midprices))
        return result
    else:
        if len(candles) < period:
            return np.nan
        return (np.max(high[-period:]) + np.min(low[-period:])) / 2



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, same_length, slice_candles


@njit
def _pvi_fast(source: np.ndarray, volume: np.ndarray) -> np.ndarray:
    """
    Numba optimized PVI calculation
    """
    pvi = np.zeros_like(source)
    pvi[0] = 1000  # Starting value

    for i in range(1, len(source)):
        if volume[i] > volume[i-1]:
            pvi[i] = pvi[i-1] * (1 + (source[i] - source[i-1]) / source[i-1])
        else:
            pvi[i] = pvi[i-1]

    return pvi


def pvi(candles: np.ndarray, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    PVI - Positive Volume Index

    The Positive Volume Index (PVI) focuses on days when volume increases from the previous day.
    The premise behind the PVI is that price changes accompanied by increased volume are more significant.

    :param candles: np.ndarray
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    res = _pvi_fast(np.ascontiguousarray(source), np.ascontiguousarray(candles[:, 5]))

    return same_length(candles, res) if sequential else res[-1]



from collections import namedtuple

import numpy as np
from scipy.signal import argrelextrema

from jesse.helpers import np_ffill, slice_candles

EXTREMA = namedtuple('EXTREMA', ['is_min', 'is_max', 'last_min', 'last_max'])


def minmax(candles: np.ndarray, order: int = 3, sequential: bool = False) -> EXTREMA:
    """
    minmax - Get extrema

    :param candles: np.ndarray
    :param order: int - default = 3
    :param sequential: bool - default: False

    :return: EXTREMA(min, max, last_min, last_max)
    """
    candles = slice_candles(candles, sequential)

    low = candles[:, 4]
    high = candles[:, 3]

    minimaIdxs = argrelextrema(low, np.less, order=order, axis=0)
    maximaIdxs = argrelextrema(high, np.greater, order=order, axis=0)

    is_min = np.full_like(low, np.nan)
    is_max = np.full_like(high, np.nan)

    # set the extremas with the matching price
    is_min[minimaIdxs] = low[minimaIdxs]
    is_max[maximaIdxs] = high[maximaIdxs]

    # forward fill Nan values to get the last extrema
    last_min = np_ffill(is_min)
    last_max = np_ffill(is_max)

    if sequential:
        return EXTREMA(is_min, is_max, last_min, last_max)
    else:
        return EXTREMA(is_min[-(order+1)], is_max[-(order+1)], last_min[-1], last_max[-1])



import json
import os
import numpy as np
import peewee
from fastapi.responses import FileResponse
import jesse.helpers as jh
from jesse.info import live_trading_exchanges, backtesting_exchanges


def get_candles(exchange: str, symbol: str, timeframe: str):
    from jesse.services.db import database
    database.open_connection()

    from jesse.services.candle import generate_candle_from_one_minutes
    from jesse.models.Candle import fetch_candles_from_db

    if 'hyperliquid' not in exchange.lower():
        symbol = symbol.upper()

    # fetch the current value for warmup_candles from the database
    from jesse.models.Option import Option
    o = Option.get(Option.type == 'config')
    db_config = json.loads(o.json)
    warmup_candles_num = db_config['live']['warm_up_candles']

    one_min_count = jh.timeframe_to_one_minutes(timeframe)
    finish_date = jh.now(force_fresh=True)
    start_date = jh.get_candle_start_timestamp_based_on_timeframe(timeframe, warmup_candles_num)

    # fetch value of generate_candles_from_1m fresh from the database
    o = Option.get(Option.type == 'config')
    generate_candles_from_1m: bool = json.loads(o.json)['live']['generate_candles_from_1m']

    # fetch 1m candles from database
    if generate_candles_from_1m:
        timeframe_to_fetch = '1m'
    else:
        timeframe_to_fetch = timeframe

    candles = np.array(
        fetch_candles_from_db(exchange, symbol, timeframe_to_fetch, start_date, finish_date)
    )

    # if there are no candles in the database, return []
    if candles.size == 0:
        database.close_connection()
        return []

    if generate_candles_from_1m:
        # leave out first candles until the timestamp of the first candle is the beginning of the timeframe
        timeframe_duration = one_min_count * 60_000
        while candles[0][0] % timeframe_duration != 0:
            candles = candles[1:]

        # generate bigger candles from 1m candles
        if timeframe != '1m':
            generated_candles = []
            for i in range(len(candles)):
                if (i + 1) % one_min_count == 0:
                    bigger_candle = generate_candle_from_one_minutes(
                        timeframe,
                        candles[(i - (one_min_count - 1)):(i + 1)],
                        True
                    )
                    generated_candles.append(bigger_candle)

            candles = generated_candles

    database.close_connection()

    return [
        {
            'time': int(c[0] / 1000),
            'open': c[1],
            'close': c[2],
            'high': c[3],
            'low': c[4],
            'volume': c[5],
        } for c in candles
    ]


def get_config(client_config: dict, has_live=False) -> dict:
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.Option import Option

    try:
        o = Option.get(Option.type == 'config')

        # merge it with client's config (because it could include new keys added),
        # update it in the database, and then return it
        data = jh.merge_dicts(client_config, json.loads(o.json))

        # make sure the list of BACKTEST exchanges is up to date
        for k in list(data['backtest']['exchanges'].keys()):
            if k not in backtesting_exchanges:
                del data['backtest']['exchanges'][k]

        # make sure the list of LIVE exchanges is up to date
        if has_live:
            for k in list(data['live']['exchanges'].keys()):
                if k not in live_trading_exchanges:
                    del data['live']['exchanges'][k]

        o.updated_at = jh.now(True)
        o.save()
    except peewee.DoesNotExist:
        # if not found, that means it's the first time. Store in the DB and
        # then return what was sent from the client side without changing it
        o = Option({
            'id': jh.generate_unique_id(),
            'updated_at': jh.now(True),
            'type': 'config',
            'json': json.dumps(client_config)
        })
        o.save(force_insert=True)

        data = client_config

    database.close_connection()

    return {
        'data': data
    }


def update_config(client_config: dict):
    from jesse.services.db import database
    database.open_connection()

    from jesse.models.Option import Option

    # at this point there must already be one option record for "config" existing, so:
    o = Option.get(Option.type == 'config')

    o.json = json.dumps(client_config)
    o.updated_at = jh.now(True)

    o.save()

    database.close_connection()


def download_file(mode: str, file_type: str, session_id: str = None):
    if mode == 'backtest' and file_type == 'log':
        path = f'storage/logs/backtest-mode/{session_id}.txt'
        filename = f'backtest-{session_id}.txt'
    elif mode == 'backtest' and file_type == 'csv':
        path = f'storage/csv/{session_id}.csv'
        filename = f'backtest-{session_id}.csv'
    elif mode == 'backtest' and file_type == 'json':
        path = f'storage/json/{session_id}.json'
        filename = f'backtest-{session_id}.json'
    elif mode == 'backtest' and file_type == 'full-reports':
        path = f'storage/full-reports/{session_id}.html'
        filename = f'backtest-{session_id}.html'
    elif mode == 'backtest' and file_type == 'tradingview':
        path = f'storage/trading-view-pine-editor/{session_id}.txt'
        filename = f'backtest-{session_id}.txt'
    elif mode == 'optimize' and file_type == 'log':
        path = f'storage/logs/optimize-mode.txt'
        # filename should be "optimize-" + current timestamp
        filename = f'optimize-{jh.timestamp_to_date(jh.now(True))}.txt'
    else:
        raise Exception(f'Unknown file type: {file_type} or mode: {mode}')

    return FileResponse(path=path, filename=filename, media_type='application/octet-stream')


def get_backtest_logs(session_id: str):
    path = f"storage/logs/backtest-mode/{session_id}.txt"

    if not os.path.exists(path):
        return None

    with open(path, "r") as f:
        content = f.read()

    return jh.compressed_response(content)


def download_backtest_log(session_id: str):
    """
    Returns the log file for a specific backtest session as a downloadable file
    """
    path = f'storage/logs/backtest-mode/{session_id}.txt'
    
    if not os.path.exists(path):
        raise Exception('Log file not found')
        
    filename = f'backtest-{session_id}.txt'
    return FileResponse(
        path=path,
        filename=filename,
        media_type='text/plain'
    )



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


def mcginley_dynamic(candles: np.ndarray, period: int = 10, k: float = 0.6, source_type: str = "close",
                     sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    McGinley Dynamic

    :param candles: np.ndarray
    :param period: int - default: 10
    :param k: float - default: 0.6
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    mg = md_fast(source, k, period)

    return mg if sequential else mg[-1]


@njit(cache=True)
def md_fast(source, k, period):
    mg = np.full_like(source, np.nan)
    for i in range(source.size):
        if i == 0:
            mg[i] = source[i]
        else:
            mg[i] = mg[i - 1] + ((source[i] - mg[i - 1]) / max([(k * period * ((source[i] / mg[i - 1]) ** 4)), 1]))

    return mg



from collections import namedtuple

import numpy as np
from .ema import ema

SupportResistanceWithBreaks = namedtuple('SupportResistanceWithBreaks', ['support', 'resistance', 'red_break', 'green_break', 'bear_wick', 'bull_wick'])


def support_resistance_with_breaks(candles: np.ndarray, left_bars: int = 15, right_bars: int = 15, vol_threshold: int = 20) -> SupportResistanceWithBreaks:
    """
    support_resistance_with_breaks

    @author LuxAlgo
    credits: https://www.tradingview.com/script/JDFoWQbL-Support-and-Resistance-Levels-with-Breaks-LuxAlgo

    :param candles: np.ndarray
    :param left_bars: int - default: 15
    :param right_bars: int - default: 15
    :param vol_threshold: int - default: 20
    :return: SupportResistanceWithBreaks(support, resistance, red_break, green_break, bear_wick, bull_wick)
    """
    resistance = _resistance(candles[:, 3], left_bars, right_bars)
    support = _support(candles[:, 4], left_bars, right_bars)

    short = ema(candles[:, 5], 5)
    long = ema(candles[:, 5], 10)
    osc = 100 * (short - long) / long

    last_candles = candles[0]

    red_break = True if last_candles[2] < support and not abs(
        last_candles[1] - last_candles[2]) < abs(last_candles[1] - last_candles[3]) and osc > vol_threshold else False
    green_break = True if last_candles[2] > resistance and abs(
        last_candles[1] - last_candles[4]) > abs(last_candles[1] - last_candles[2]) and osc > vol_threshold else False

    bull_wick = True if last_candles[2] > resistance and abs(last_candles[1] - last_candles[4]) > abs(last_candles[1] - last_candles[2]) else False
    bear_wick = True if last_candles[2] < support and abs(last_candles[1] - last_candles[2]) < abs(last_candles[1] - last_candles[3]) else False

    return SupportResistanceWithBreaks(support, resistance, red_break, green_break, bear_wick, bull_wick)


def _resistance(source, left_bars, right_bars):
    pivot_highs = [None] * len(source)  # Initialize result list with None

    for i in range(left_bars, len(source) - right_bars):
        is_pivot_high = True

        # Check left bars for higher high
        for j in range(1, left_bars + 1):
            if source[i] <= source[i - j]:
                is_pivot_high = False
                break

        # Check right bars for higher high
        if is_pivot_high:
            for j in range(1, right_bars + 1):
                if source[i] <= source[i + j]:
                    is_pivot_high = False
                    break

        if is_pivot_high:
            is_pivot_high = source[i]

        pivot_highs[i] = is_pivot_high

    next_valid = None
    first_value = None
    for i in range(len(pivot_highs)):
        if pivot_highs[i] is False:
            pivot_highs[i] = next_valid
        elif pivot_highs[i] is not None:  # Update next_valid if it's not False or None
            next_valid = pivot_highs[i]
            first_value = i if first_value is None else first_value

    pivot_highs[:first_value - 1] = [pivot_highs[first_value]] * len(pivot_highs[:first_value - 1])
    pivot_highs[-right_bars:] = [pivot_highs[-right_bars - 1]] * len(pivot_highs[-right_bars:])

    return pivot_highs[-1]


def _support(source, left_bars, right_bars):
    pivot_lows = [None] * len(source)  # Initialize result list with None

    for i in range(left_bars, len(source) - right_bars):
        is_pivot_low = True

        # Check left bars for lower low
        for j in range(1, left_bars + 1):
            if source[i] >= source[i - j]:
                is_pivot_low = False
                break

        # Check right bars for lower low
        if is_pivot_low:
            for j in range(1, right_bars + 1):
                if source[i] >= source[i + j]:
                    is_pivot_low = False
                    break

        if is_pivot_low:
            is_pivot_low = source[i]

        pivot_lows[i] = is_pivot_low

    next_valid = None
    first_value = None
    for i in range(len(pivot_lows)):
        if pivot_lows[i] is False:
            pivot_lows[i] = next_valid
        elif pivot_lows[i] is not None:  # Update next_valid if it's not False or None
            next_valid = pivot_lows[i]
            first_value = i if first_value is None else first_value

    pivot_lows[:first_value - 1] = [pivot_lows[first_value]] * len(pivot_lows[:first_value - 1])
    pivot_lows[-right_bars:] = [pivot_lows[-right_bars - 1]] * len(pivot_lows[-right_bars:])

    return pivot_lows[-1]



from multiprocessing import cpu_count
from typing import Dict, List, Tuple
import arrow
import jesse.helpers as jh
from jesse.modes.backtest_mode import load_candles
from jesse.services.validators import validate_routes
from jesse.store import store
from .Optimize import Optimizer
from jesse.services.failure import register_custom_exception_handler
from jesse.routes import router
from jesse.models.OptimizationSession import store_optimization_session, get_optimization_session_by_id, update_optimization_session_status, update_optimization_session_state


def run(
        session_id: str,
        user_config: dict,
        exchange: str,
        routes: List[Dict[str, str]],
        data_routes: List[Dict[str, str]],
        training_start_date: str,
        training_finish_date: str,
        testing_start_date: str,
        testing_finish_date: str,
        optimal_total: int,
        fast_mode: bool,
        cpu_cores: int,
        state: dict,
) -> None:
    if jh.python_version() == (3, 13):
        raise ValueError(
            'Optimization is not supported on Python 3.13. The "Ray" library used for optimization does not support Python 3.13 yet. Please use Python 3.12 or lower.')

    from jesse.config import config, set_config
    config['app']['trading_mode'] = 'optimize'

    # validate cpu_cores
    if cpu_cores < 1:
        raise ValueError('cpu_cores must be an integer value greater than 0. Please check your settings page for optimization.')
    # get the max number of cores
    max_cpu_cores = cpu_count()
    if cpu_cores > max_cpu_cores:
        raise ValueError(f'cpu_cores must be less than or equal to {max_cpu_cores} which is the number of cores on your machine.')

    # inject config
    set_config(user_config)
    # add exchange to routes
    for r in routes:
        r['exchange'] = exchange
    for r in data_routes:
        r['exchange'] = exchange
    # set routes
    router.initiate(routes, data_routes)
    store.app.set_session_id(session_id)
    register_custom_exception_handler()
    # validate routes
    validate_routes(router)

    # load historical candles
    training_warmup_candles, training_candles, testing_warmup_candles, testing_candles = _get_training_and_testing_candles(
        training_start_date,
        training_finish_date,
        testing_start_date,
        testing_finish_date
    )

    # Check if we're resuming an existing session
    existing_session = get_optimization_session_by_id(session_id)

    if existing_session:
        # Session exists, update it for resuming
        update_optimization_session_status(session_id, 'running')
        update_optimization_session_state(session_id, state)

        if jh.is_debugging():
            jh.debug(f"Resuming existing optimization session with ID: {session_id}")
    else:
        # Session doesn't exist, create a new one
        store_optimization_session(
            id=session_id,
            status='running'
        )
        update_optimization_session_state(session_id, state)

        if jh.is_debugging():
            jh.debug(f"Created new optimization session with ID: {session_id}")

    optimizer = Optimizer(
        session_id,
        user_config,
        training_warmup_candles,
        training_candles,
        testing_warmup_candles,
        testing_candles,
        fast_mode,
        optimal_total,
        cpu_cores
    )

    optimizer.run()


def _get_training_and_testing_candles(
        training_start_date: str,
        training_finish_date: str,
        testing_start_date: str,
        testing_finish_date: str
) -> Tuple[dict, dict, dict, dict]:
    training_start_date_timestamp = jh.arrow_to_timestamp(arrow.get(training_start_date, 'YYYY-MM-DD'))
    training_finish_date_timestamp = jh.arrow_to_timestamp(arrow.get(training_finish_date, 'YYYY-MM-DD'))
    testing_start_date_timestamp = jh.arrow_to_timestamp(arrow.get(testing_start_date, 'YYYY-MM-DD'))
    testing_finish_date_timestamp = jh.arrow_to_timestamp(arrow.get(testing_finish_date, 'YYYY-MM-DD'))

    # fetch training candles
    training_warmup_candles, training_candles = load_candles(training_start_date_timestamp, training_finish_date_timestamp)
    # fetch testing candles
    testing_warmup_candles, testing_candles = load_candles(testing_start_date_timestamp, testing_finish_date_timestamp)

    return training_warmup_candles, training_candles, testing_warmup_candles, testing_candles



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

Volume = namedtuple("Volume", ["volume", "ma"])


def volume(
    candles: np.ndarray,
    period: int = 20,
    sequential: bool = False
) -> Volume:
    """
    Volume with Moving Average

    :param candles: np.ndarray
    :param period: int - default: 20
    :param sequential: bool - default: False

    :return: Volume(volume, ma)
    """
    candles = slice_candles(candles, sequential)

    volume_data = candles[:, 5]
    if len(volume_data) < period:
        volume_ma = np.full(len(volume_data), np.nan)
    else:
        volume_ma = np.concatenate((np.full(period - 1, np.nan), np.convolve(volume_data, np.ones(period) / period, mode='valid')))

    if sequential:
        return Volume(volume_data, volume_ma)
    else:
        return Volume(volume_data[-1], volume_ma[-1])



class EmptyPosition(Exception):
    pass


class OpenPositionError(Exception):
    pass


class OrderNotAllowed(Exception):
    pass


class ConflictingRules(Exception):
    pass


class InvalidStrategy(Exception):
    pass


class CandleNotFoundInDatabase(Exception):
    pass


class CandleNotFoundInExchange(Exception):
    pass


class SymbolNotFound(Exception):
    pass


class RouteNotFound(Exception):
    def __init__(self, symbol, timeframe):
        message = f"Data route is required but missing: symbol='{symbol}', timeframe='{timeframe}'"
        super().__init__(message)


class InvalidRoutes(Exception):
    pass


class ExchangeInMaintenance(Exception):
    pass


class ExchangeNotResponding(Exception):
    pass


class ExchangeRejectedOrder(Exception):
    pass

class ExchangeRejectedLeverageNumber(Exception):
    pass


class ExchangeOrderNotFound(Exception):
    pass


class InvalidShape(Exception):
    pass


class InvalidConfig(Exception):
    pass


class InvalidTimeframe(Exception):
    pass


class InvalidSymbol(Exception):
    pass


class NegativeBalance(Exception):
    pass


class InsufficientMargin(Exception):
    pass


class InsufficientBalance(Exception):
    pass


class Termination(Exception):
    pass


class InvalidExchangeApiKeys(Exception):
    pass


class ExchangeError(Exception):
    pass


class NotSupportedError(Exception):
    pass


class CandlesNotFound(Exception):
    pass


class InvalidDateRange(Exception):
    pass



from .ApexProMain import ApexProMain
from jesse.enums import exchanges


class ApexProPerpetual(ApexProMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.APEX_PRO_PERPETUAL,
            rest_endpoint='https://pro.apex.exchange/api/v2'
        )



from datetime import datetime

import arrow
import numpy as np

import jesse.helpers as jh
from jesse.config import config
from jesse.exceptions import CandleNotFoundInDatabase
from jesse.models import Candle
from jesse.services.cache import cache
from jesse.services.candle import generate_candle_from_one_minutes
from jesse.store import store


def load_required_candles(exchange: str, symbol: str, start_date_str: str, finish_date_str: str) -> np.ndarray:
    """
    loads initial candles that required before executing strategies.
    210 for the biggest timeframe and more for the rest
    """
    start_date = jh.arrow_to_timestamp(arrow.get(start_date_str, 'YYYY-MM-DD'))
    finish_date = jh.arrow_to_timestamp(arrow.get(finish_date_str, 'YYYY-MM-DD')) - 60000

    # validate
    if start_date == finish_date:
        raise ValueError('start_date and finish_date cannot be the same.')
    if start_date > finish_date:
        raise ValueError('start_date cannot be bigger than finish_date.')
    if finish_date > arrow.utcnow().int_timestamp * 1000:
        raise ValueError('Can\'t backtest the future!')

    max_timeframe = jh.max_timeframe(config['app']['considering_timeframes'])
    short_candles_count = jh.get_config('env.data.warmup_candles_num', 210) * jh.timeframe_to_one_minutes(max_timeframe)
    pre_finish_date = start_date - 60_000
    pre_start_date = pre_finish_date - short_candles_count * 60_000
    # make sure starting from the beginning of the day instead
    pre_start_date = jh.timestamp_to_arrow(pre_start_date).floor('day').int_timestamp * 1000
    # update candles_count to count from the beginning of the day instead
    short_candles_count = int((pre_finish_date - pre_start_date) / 60_000)

    key = jh.key(exchange, symbol)
    cache_key = f'{jh.timestamp_to_date(pre_start_date)}-{jh.timestamp_to_date(pre_finish_date)}-{key}'
    cached_value = cache.get_value(cache_key)

    # if cache exists
    if cached_value:
        candles_tuple = cached_value
    # not cached, get and cache for later calls in the next 5 minutes
    else:
        # fetch from database
        candles_tuple = tuple(
            Candle.select(
                Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
                Candle.volume
            ).where(
                Candle.exchange == exchange,
                Candle.symbol == symbol,
                Candle.timeframe == '1m' or Candle.timeframe.is_null(),
                Candle.timestamp.between(pre_start_date, pre_finish_date)
            ).order_by(Candle.timestamp.asc()).tuples()
        )

        # cache it for near future calls
        cache.set_value(cache_key, candles_tuple, expire_seconds=60 * 60 * 24 * 7)

    candles = np.array(candles_tuple)

    if len(candles) < short_candles_count + 1:
        first_existing_candle = tuple(
            Candle.select(Candle.timestamp).where(
                Candle.exchange == exchange,
                Candle.symbol == symbol,
                Candle.timeframe == '1m' or Candle.timeframe.is_null()
            ).order_by(Candle.timestamp.asc()).limit(1).tuples()
        )

        if not len(first_existing_candle):
            raise CandleNotFoundInDatabase(
                f'No candle for "{exchange}" "{symbol}" is present in the database. Try importing candles.'
            )

        first_existing_candle = first_existing_candle[0][0]

        last_existing_candle = tuple(
            Candle.select(Candle.timestamp).where(
                Candle.exchange == exchange,
                Candle.symbol == symbol,
                Candle.timeframe == '1m' or Candle.timeframe.is_null()
            ).order_by(Candle.timestamp.desc()).limit(1).tuples()
        )[0][0]

        first_backtestable_timestamp = first_existing_candle + (pre_finish_date - pre_start_date) + (60_000 * 1440)
        first_backtestable_datetime = datetime.fromtimestamp(first_backtestable_timestamp/1000)

        # if first backtestable timestamp is in the future, that means we have some but not enough candles
        if first_backtestable_timestamp > jh.today_to_timestamp():
            raise CandleNotFoundInDatabase(
                f'Not enough candle for {exchange} {symbol} is present in the database. Jesse requires "210 * biggest_timeframe" warm-up candles. '
                'Try importing more candles from an earlier date.'
            )

        raise CandleNotFoundInDatabase(
            f'Not enough candles for {exchange} {symbol} exists to run backtest from {start_date_str} => {finish_date_str} '
            f'(first possible date is {first_backtestable_datetime.date()}). \n'

            f'Are you considering the warmup candles? For more info please read:\n https://jesse.trade/help/faq/i-imported-candles-but-keep-getting-not-enough-candles'
        )

    return candles


def inject_required_candles_to_store(candles: np.ndarray, exchange: str, symbol: str) -> None:
    """
    generate and add required candles to the candle store
    """
    # batch add 1m candles:
    store.candles.batch_add_candle(candles, exchange, symbol, '1m', with_generation=False)

    # loop to generate, and add candles (without execution)
    for i in range(len(candles)):
        for timeframe in config['app']['considering_timeframes']:
            # skip 1m. already added
            if timeframe == '1m':
                continue

            num = jh.timeframe_to_one_minutes(timeframe)

            if (i + 1) % num == 0:
                generated_candle = generate_candle_from_one_minutes(
                    timeframe,
                    candles[(i - (num - 1)):(i + 1)],
                    True
                )

                store.candles.add_candle(
                    generated_candle,
                    exchange,
                    symbol,
                    timeframe,
                    with_execution=False,
                    with_generation=False
                )



from typing import Union

import numpy as np
from jesse.helpers import get_candle_source, slice_candles


def rocp(candles: np.ndarray, period: int = 10, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    ROCP - Rate of change Percentage: (price-prevPrice)/prevPrice

    :param candles: np.ndarray
    :param period: int - default: 10
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    res = np.full(source.shape, np.nan, dtype=float)
    if len(source) > period:
        res[period:] = (source[period:] - source[:-period]) / source[:-period]

    return res if sequential else res[-1]



import warnings
from typing import Optional, Dict, Set
import click
import pkg_resources
from fastapi import BackgroundTasks, Query, Header
from fastapi.responses import JSONResponse, FileResponse
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
from starlette.websockets import WebSocket, WebSocketDisconnect
from jesse.services.multiprocessing import process_manager
from jesse.services.web import fastapi_app
from jesse.services.redis import async_redis, async_publish, sync_publish
from jesse.services.web import fastapi_app, BacktestRequestJson, ImportCandlesRequestJson, CancelRequestJson, \
    LoginRequestJson, ConfigRequestJson, LoginJesseTradeRequestJson, NewStrategyRequestJson, FeedbackRequestJson, \
    ReportExceptionRequestJson, OptimizationRequestJson, StoreExchangeApiKeyRequestJson, \
    DeleteExchangeApiKeyRequestJson, StoreNotificationApiKeyRequestJson, DeleteNotificationApiKeyRequestJson, \
    ExchangeSupportedSymbolsRequestJson, SaveStrategyRequestJson, GetStrategyRequestJson, DeleteStrategyRequestJson, \
    DeleteCandlesRequestJson
from jesse.services import auth as authenticator

from jesse.services.ws_manager import ws_manager
import uvicorn
import jesse.helpers as jh
import time

# variable to know if the live trade plugin is installed
HAS_LIVE_TRADE_PLUGIN = True
try:
    import jesse_live
except ModuleNotFoundError:
    HAS_LIVE_TRADE_PLUGIN = False

# to silent stupid pandas warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# get the jesse directory
JESSE_DIR = pkg_resources.resource_filename(__name__, '')


# load homepage
@fastapi_app.get("/")
async def index():
    return FileResponse(f"{JESSE_DIR}/static/index.html")


@fastapi_app.post("/terminate-all")
async def terminate_all(authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.flush()
    return JSONResponse({'message': 'terminating all tasks...'})


@fastapi_app.post("/shutdown")
async def shutdown(background_tasks: BackgroundTasks, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    background_tasks.add_task(jh.terminate_app)
    return JSONResponse({'message': 'Shutting down...'})


@fastapi_app.post("/auth")
def auth(json_request: LoginRequestJson):
    return authenticator.password_to_token(json_request.password)


@fastapi_app.post("/make-strategy")
def make_strategy(json_request: NewStrategyRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_handler
    return strategy_handler.generate(json_request.name)


@fastapi_app.post("/feedback")
def feedback(json_request: FeedbackRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    return jesse_trade.feedback(json_request.description, json_request.email)


@fastapi_app.post("/report-exception")
def report_exception(json_request: ReportExceptionRequestJson,
                     authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    return jesse_trade.report_exception(
        json_request.description,
        json_request.traceback,
        json_request.mode,
        json_request.attach_logs,
        json_request.session_id,
        json_request.email,
        has_live=HAS_LIVE_TRADE_PLUGIN
    )


@fastapi_app.post("/get-config")
def get_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import get_config as gc

    return JSONResponse({
        'data': gc(json_request.current_config, has_live=HAS_LIVE_TRADE_PLUGIN)
    }, status_code=200)


@fastapi_app.post("/update-config")
def update_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import update_config as uc

    uc(json_request.current_config)

    return JSONResponse({'message': 'Updated configurations successfully'}, status_code=200)


@fastapi_app.post("/clear-candles-database-cache")
def clear_candles_database_cache(authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.cache import cache
    cache.flush()

    return JSONResponse({
        'status': 'success',
        'message': 'Candles database cache cleared successfully',
    }, status_code=200)


@fastapi_app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket, token: str = Query(...)):
    from jesse.services.env import ENV_VALUES

    if not authenticator.is_valid_token(token):
        return

    # Use connection manager to handle this websocket
    connection_id = str(id(websocket))
    print(jh.color(f"=> WebSocket {connection_id} connecting", 'yellow'))
    
    await ws_manager.connect(websocket)
    channel_pattern = f"{ENV_VALUES['APP_PORT']}:channel:*"
    
    # Start Redis listener if not already started
    await ws_manager.start_redis_listener(channel_pattern)
    
    try:
        # Keep the connection alive
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        print(jh.color(f"WebSocket {connection_id} disconnected", 'yellow'))
        ws_manager.disconnect(websocket)
        # Optionally stop Redis listener if no more clients
        await ws_manager.stop_redis_listener()
    except Exception as e:
        print(jh.color(f"WebSocket error: {str(e)}", 'red'))
        ws_manager.disconnect(websocket)
        await ws_manager.stop_redis_listener()

# create a Click group
@click.group()
@click.version_option(pkg_resources.get_distribution("jesse").version)
def cli() -> None:
    pass


@cli.command()
@click.option(
    '--strict/--no-strict', default=True,
    help='Default is the strict mode which will raise an exception if the values for license is not set.'
)
def install_live(strict: bool) -> None:
    from jesse.services.installer import install
    install(is_live_plugin_already_installed=jh.has_live_trade_plugin(), strict=strict)


@cli.command()
def run() -> None:
    # Display welcome message
    welcome_message = """
     ██╗███████╗███████╗███████╗███████╗
     ██║██╔════╝██╔════╝██╔════╝██╔════╝
     ██║█████╗  ███████╗███████╗█████╗  
██   ██║██╔══╝  ╚════██║╚════██║██╔══╝  
╚█████╔╝███████╗███████║███████║███████╗
 ╚════╝ ╚══════╝╚══════╝╚══════╝╚══════╝
                                        
    """
    version = pkg_resources.get_distribution("jesse").version
    print(welcome_message)
    print(f"Main Framework Version: {version}")
    
    # Check if jesse-live is installed and display its version
    if jh.has_live_trade_plugin():
        try:
            from jesse_live.version import __version__ as live_version
            print(f"Live Plugin Version: {live_version}")
        except ImportError:
            pass
    
    jh.validate_cwd()

    print("")

    # run all the db migrations
    from jesse.services.migrator import run as run_migrations
    import peewee
    try:
        run_migrations()
    except peewee.OperationalError:
        sleep_seconds = 10
        print(f"Database wasn't ready. Sleep for {sleep_seconds} seconds and try again.")
        time.sleep(sleep_seconds)
        run_migrations()

    # read port from .env file, if not found, use default
    from jesse.services.env import ENV_VALUES
    if 'APP_PORT' in ENV_VALUES:
        port = int(ENV_VALUES['APP_PORT'])
    else:
        port = 9000

    if 'APP_HOST' in ENV_VALUES:
        host = ENV_VALUES['APP_HOST']
    else:
        host = "0.0.0.0"

    # run the main application
    process_manager.flush()
    uvicorn.run(fastapi_app, host=host, port=port, log_level="info")


@fastapi_app.on_event("shutdown")
def shutdown_event():
    from jesse.services.db import database
    database.close_connection()


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Routes
# # # # # # # # # # # # # # # # # # # # # # # # # # # #
from jesse.controllers.websocket_controller import router as websocket_router
from jesse.controllers.optimization_controller import router as optimization_router
from jesse.controllers.exchange_controller import router as exchange_router
from jesse.controllers.backtest_controller import router as backtest_router
from jesse.controllers.candles_controller import router as candles_router
from jesse.controllers.strategy_controller import router as strategy_router
from jesse.controllers.auth_controller import router as auth_router
from jesse.controllers.config_controller import router as config_router
from jesse.controllers.notification_controller import router as notification_router
from jesse.controllers.system_controller import router as system_router
from jesse.controllers.file_controller import router as file_router

# register routers
fastapi_app.include_router(websocket_router)
fastapi_app.include_router(optimization_router)
fastapi_app.include_router(exchange_router)
fastapi_app.include_router(backtest_router)
fastapi_app.include_router(candles_router)
fastapi_app.include_router(strategy_router)
fastapi_app.include_router(auth_router)
fastapi_app.include_router(config_router)
fastapi_app.include_router(notification_router)
fastapi_app.include_router(system_router)
fastapi_app.include_router(file_router)


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Live Trade Plugin
# # # # # # # # # # # # # # # # # # # # # # # # # # # #
if jh.has_live_trade_plugin():
    from jesse.controllers.live_controller import router as live_router
    fastapi_app.include_router(live_router)


# # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Static Files (Must be loaded at the end to prevent overlapping with API endpoints)
# # # # # # # # # # # # # # # # # # # # # # # # # # # #
fastapi_app.mount("/", StaticFiles(directory=f"{JESSE_DIR}/static"), name="static")



from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.factories import fake_order
from jesse.store import store
from jesse.routes import router
import jesse.helpers as jh


def set_up():
    reset_config()
    config['app']['trading_exchanges'] = [exchanges.SANDBOX]
    config['app']['trading_symbols'] = ['BTC-USD']
    routes = [
        {'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD', 'timeframe': '1m', 'strategy': 'TestVanillaStrategy'}
    ]
    router.initiate(routes)


def test_add_new_order():
    set_up()

    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.storage['Sandbox-BTC-USD'] == [o1, o2]


def test_state_order_count():
    set_up()

    assert store.orders.count(exchanges.SANDBOX, 'BTC-USD') == 0
    store.orders.add_order(fake_order())
    assert store.orders.count(exchanges.SANDBOX, 'BTC-USD') == 1
    store.orders.add_order(fake_order())
    assert store.orders.count(exchanges.SANDBOX, 'BTC-USD') == 2


def test_state_order_get_order_by_id():
    set_up()

    o0 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})

    store.orders.get_order_by_id(exchanges.SANDBOX, 'BTC-USD', o2.id)

    # return None if it does not exist
    assert store.orders.get_order_by_id(exchanges.SANDBOX, 'BTC-USD', o0.id) is None

    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.get_order_by_id(exchanges.SANDBOX, 'BTC-USD', o2.id) == o2


def test_state_order_get_orders():
    set_up()

    o1 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    o2 = fake_order({'exchange': exchanges.SANDBOX, 'symbol': 'BTC-USD'})
    store.orders.add_order(o1)
    store.orders.add_order(o2)
    assert store.orders.get_orders(exchanges.SANDBOX, 'BTC-USD') == [o1, o2]



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestTerminate(Strategy):
    def __init__(self):
        super().__init__()
        self.__before_terminate_called = False

    def should_long(self) -> bool:
        return self.price == 10

    def go_long(self) -> None:
        self.buy = 1, self.price

    def should_cancel_entry(self):
        return False

    def before_terminate(self):
        self.__before_terminate_called = True

    def terminate(self):
        assert self.__before_terminate_called is True

        # change a value in store so we can assert the change in the original unit test
        from jesse.store import store
        store.app.starting_time = 1



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import same_length, slice_candles


@njit
def _qstick_fast(open_prices: np.ndarray, close_prices: np.ndarray, period: int) -> np.ndarray:
    """
    Calculate QStick values using Numba for optimization
    """
    # Pre-allocate output array
    qstick_values = np.zeros_like(open_prices, dtype=np.float64)

    # Calculate close-open difference
    diff = close_prices - open_prices

    # Calculate moving average of the difference
    for i in range(period - 1, len(diff)):
        qstick_values[i] = np.mean(diff[i - period + 1:i + 1])

    return qstick_values


def qstick(candles: np.ndarray, period: int = 5, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    QStick - Moving average of the difference between closing and opening prices

    :param candles: np.ndarray
    :param period: int - default: 5
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    res = _qstick_fast(
        np.ascontiguousarray(candles[:, 1]),  # open
        np.ascontiguousarray(candles[:, 2]),  # close
        period
    )

    return same_length(candles, res) if sequential else res[-1]



from jesse.strategies import Strategy
import jesse.helpers as jh
from jesse import utils


class TestPositionTotalCostProperty(Strategy):
    def before(self) -> None:
        if self.price == 20:
            if self.exchange_type == 'futures':
                assert self.position.entry_price * self.position.qty == 100
                assert self.leverage == 2
                assert self.position.total_cost == 50
            elif self.exchange_type == 'spot':
                assert self.position.entry_price * self.position.qty == 100
                assert self.leverage == 1
                assert self.position.total_cost == 100

    def should_long(self) -> bool:
        return self.price == 10

    def go_long(self) -> None:
        self.buy = 10, self.price

    def should_cancel_entry(self):
        return False



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def stddev(candles: np.ndarray, period: int = 5, nbdev: float = 1, source_type: str = "close",
           sequential: bool = False) -> Union[float, np.ndarray]:
    """
    STDDEV - Standard Deviation

    :param candles: np.ndarray
    :param period: int - default: 5
    :param nbdev: float - default: 1
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    n = len(source)
    # Initialize result array with nan values
    result = np.full(n, np.nan, dtype=float)

    if n < period:
        # Not enough data for full period, result remains as nans
        output = result
    else:
        # Create rolling windows using numpy's sliding_window_view
        windows = np.lib.stride_tricks.sliding_window_view(source, window_shape=period)
        # Compute standard deviation over the rolling windows using population std (ddof=0) and multiply by nbdev
        rolling_std = np.std(windows, axis=1, ddof=0) * nbdev
        # Fill the result array from index 'period - 1' onward with the computed rolling std
        result[period - 1:] = rolling_std
        output = result

    return output if sequential else output[-1]



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, slice_candles


def trima(candles: np.ndarray, period: int = 30, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    TRIMA - Triangular Moving Average

    :param candles: np.ndarray
    :param period: int - default: 30
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    # Compute triangular weights
    if period % 2 != 0:
        mid = period // 2
        weights = np.concatenate((np.arange(1, mid + 2), np.arange(mid, 0, -1)))
    else:
        mid = period // 2
        weights = np.concatenate((np.arange(1, mid + 1), np.arange(mid, 0, -1)))
    weights_norm = weights / weights.sum()

    n = source.shape[0]
    if n < period:
        res = np.full(n, np.nan)
    else:
        conv = np.convolve(source, weights_norm, mode='valid')
        res = np.concatenate((np.full(period - 1, np.nan), conv))

    return res if sequential else res[-1]



import sys
from typing import List, Any

import jesse.helpers as jh
from jesse import exceptions
from jesse.models import Route


class RouterClass:
    def __init__(self) -> None:
        self.routes = []
        self.data_candles = [] # used in legacy code (probably will be replaced in the future with self.data_routes)
        self.data_routes = [] # same as data_candles but instead of dict, it's a Route object

    def _reset(self) -> None:
        self.routes = []
        self.data_candles = []
        self.data_routes = []

    @property
    def formatted_routes(self) -> list:
        """
        Example:
        [{'exchange': 'Binance', 'strategy': 'A1', 'symbol': 'BTC-USDT', 'timeframe': '1m'}]
        """
        return [
            {
                'exchange': r.exchange,
                'symbol': r.symbol,
                'timeframe': r.timeframe,
                'strategy': r.strategy_name,
            }
            for r in self.routes
        ]

    @property
    def formatted_data_routes(self) -> list:
        """
        Example:
        [{'exchange': 'Binance', 'symbol': 'BTC-USD', 'timeframe': '3m'}]
        """
        return [{
            'exchange': r['exchange'],
            'symbol': r['symbol'],
            'timeframe': r['timeframe']
        } for r in self.data_candles]

    @property
    def all_formatted_routes(self) -> list:
        return self.formatted_routes + self.formatted_data_routes

    def initiate(self, routes: list, data_routes: list = None):
        if data_routes is None:
            data_routes = []
        self.set_routes(routes)
        self.set_data_candles(data_routes)
        from jesse.store import store
        store.reset(force_install_routes=jh.is_unit_testing())

    def set_routes(self, routes: List[Any]) -> None:
        self._reset()

        self.routes = []

        for r in routes:
            # validate strategy that the strategy file exists (if sent as a string)
            if isinstance(r["strategy"], str):
                strategy_name = r["strategy"]
                if jh.is_unit_testing():
                    path = sys.path[0]
                    # live plugin
                    if path.endswith('jesse-live'):
                        strategies_dir = f'{sys.path[0]}/tests/strategies'
                    # main framework
                    else:
                        strategies_dir = f'{sys.path[0]}/jesse/strategies'
                    exists = jh.file_exists(f"{strategies_dir}/{strategy_name}/__init__.py")
                else:
                    exists = jh.file_exists(f'strategies/{strategy_name}/__init__.py')
            else:
                exists = True

            if not exists and isinstance(r["strategy"], str):
                raise exceptions.InvalidRoutes(
                    f'A strategy with the name of "{r["strategy"]}" could not be found.')

            self.routes.append(Route(r["exchange"], r["symbol"], r["timeframe"], r["strategy"], None))

    def set_data_routes(self, routes: List[Any]) -> None:
        self.data_routes = []
        for r in routes:
            self.data_routes.append(Route(r["exchange"], r["symbol"], r["timeframe"], None, None))

    def set_data_candles(self, data_candles: list) -> None:
        self.data_candles = data_candles
        self.set_data_routes(data_candles)


router: RouterClass = RouterClass()



from jesse.strategies import Strategy


class TestOrdersAreSortedBeforeExecution(Strategy):
    def should_long(self):
        return self.price == 10

    def go_long(self):
        self.buy = [
            (1, 10.2),
            (1, 10.3),
            (1, 10.1)
        ]

    def on_open_position(self, order) -> None:
        # the order with the lowest price should be executed first
        assert order.price == 10.1

    def update_position(self):
        if self.price == 20:
            self.liquidate()

    def should_cancel_entry(self):
        return False



from typing import Union

import numpy as np
from jesse.helpers import slice_candles
from numba import njit


def sar(candles: np.ndarray, acceleration: float = 0.02, maximum: float = 0.2, sequential: bool = False) -> Union[float, np.ndarray]:
    """
    SAR - Parabolic SAR

    :param candles: np.ndarray
    :param acceleration: float - default: 0.02
    :param maximum: float - default: 0.2
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    # Assuming candle format where index 3 is high and index 4 is low
    high = candles[:, 3]
    low = candles[:, 4]

    n = len(candles)
    if n == 0:
        return np.array([])
    if n < 2:
        return low[-1]

    # Use numba-compiled function to calculate SAR values
    sar_values = _fast_sar(high, low, acceleration, maximum, n)
    return sar_values if sequential else sar_values[-1]


@njit(cache=True)
def _fast_sar(high, low, acceleration, maximum, n):
    sar_values = np.zeros(n)
    if high[1] > high[0]:
        uptrend = True
        sar_values[0] = low[0]
        ep = high[0]  # extreme point
    else:
        uptrend = False
        sar_values[0] = high[0]
        ep = low[0]
    af = acceleration  # initial acceleration factor
    for i in range(1, n):
        prev_sar = sar_values[i - 1]
        if uptrend:
            sar_temp = prev_sar + af * (ep - prev_sar)
            if i >= 2:
                sar_temp = min(sar_temp, low[i - 1], low[i - 2])
            else:
                sar_temp = min(sar_temp, low[i - 1])
        else:
            sar_temp = prev_sar - af * (prev_sar - ep)
            if i >= 2:
                sar_temp = max(sar_temp, high[i - 1], high[i - 2])
            else:
                sar_temp = max(sar_temp, high[i - 1])

        if uptrend:
            if low[i] < sar_temp:
                sar_temp = ep
                uptrend = False
                af = acceleration  # reset acceleration factor
                ep = low[i]
            else:
                if high[i] > ep:
                    ep = high[i]
                    af = af + acceleration
                    if af > maximum:
                        af = maximum
        else:
            if high[i] > sar_temp:
                sar_temp = ep
                uptrend = True
                af = acceleration
                ep = high[i]
            else:
                if low[i] < ep:
                    ep = low[i]
                    af = af + acceleration
                    if af > maximum:
                        af = maximum

        sar_values[i] = sar_temp
    return sar_values



from typing import Union
import numpy as np
from numba import njit
from jesse.helpers import get_candle_source, same_length, slice_candles
from jesse.indicators.sma import sma


@njit
def _dpo(source, period, sma):
    # Calculate the X/2 + 1 shift
    shift = period // 2 + 1

    # Shift the price series and subtract SMA
    shifted_source = np.roll(source, shift)
    dpo = shifted_source - sma

    # First (period-1 + shift) elements will be invalid due to the rolling calculations
    dpo[:period-1+shift] = np.nan

    return dpo


def dpo(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    DPO - Detrended Price Oscillator

    Formula: Price {X/2 + 1} periods ago less the X-period simple moving average

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)
    t_sma = sma(candles, period, source_type=source_type, sequential=True)
    res = _dpo(source, period, t_sma)

    return same_length(candles, res) if sequential else res[-1]



import peewee
from jesse.services.db import database
import jesse.helpers as jh
import numpy as np


if database.is_closed():
    database.open_connection()


class Candle(peewee.Model):
    id = peewee.UUIDField(primary_key=True)
    timestamp = peewee.BigIntegerField()
    open = peewee.FloatField()
    close = peewee.FloatField()
    high = peewee.FloatField()
    low = peewee.FloatField()
    volume = peewee.FloatField()
    exchange = peewee.CharField()
    symbol = peewee.CharField()
    timeframe = peewee.CharField()

    # partial candles: 5 * 1m candle = 5m candle while 1m == partial candle
    is_partial = True

    class Meta:
        from jesse.services.db import database

        database = database.db
        indexes = (
            (('exchange', 'symbol', 'timeframe', 'timestamp'), True),
        )

    def __init__(self, attributes: dict = None, **kwargs) -> None:
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a, value in attributes.items():
            setattr(self, a, value)


# if database is open, create the table
if database.is_open():
    Candle.create_table()


# # # # # # # # # # # # # # # # # # # # # # # # # # # 
# # # # # # # # # DB FUNCTIONS # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # # # # # # # # 


def store_candle_into_db(exchange: str, symbol: str, timeframe: str, candle: np.ndarray, on_conflict='ignore') -> None:
    d = {
        'id': jh.generate_unique_id(),
        'exchange': exchange,
        'symbol': symbol,
        'timeframe': timeframe,
        'timestamp': candle[0],
        'open': candle[1],
        'high': candle[3],
        'low': candle[4],
        'close': candle[2],
        'volume': candle[5]
    }

    if on_conflict == 'ignore':
        Candle.insert(**d).on_conflict_ignore().execute()
    elif on_conflict == 'replace':
        Candle.insert(**d).on_conflict(
            conflict_target=['exchange', 'symbol', 'timeframe', 'timestamp'],
            preserve=(Candle.open, Candle.high, Candle.low, Candle.close, Candle.volume),
        ).execute()
    elif on_conflict == 'error':
        Candle.insert(**d).execute()
    else:
        raise Exception(f'Unknown on_conflict value: {on_conflict}')


def store_candles_into_db(exchange: str, symbol: str, timeframe: str, candles: np.ndarray, on_conflict='ignore') -> None:
    # make sure the number of candles is more than 0
    if len(candles) == 0:
        raise Exception(f'No candles to store for {exchange}-{symbol}-{timeframe}')

    # convert candles to list of dicts
    candles_list = []
    for candle in candles:
        d = {
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': exchange,
            'timestamp': candle[0],
            'open': candle[1],
            'high': candle[3],
            'low': candle[4],
            'close': candle[2],
            'volume': candle[5],
            'timeframe': timeframe,
        }
        candles_list.append(d)

    if on_conflict == 'ignore':
        Candle.insert_many(candles_list).on_conflict_ignore().execute()
    elif on_conflict == 'replace':
        Candle.insert_many(candles_list).on_conflict(
            conflict_target=['exchange', 'symbol', 'timeframe', 'timestamp'],
            preserve=(Candle.open, Candle.high, Candle.low, Candle.close, Candle.volume),
        ).execute()
    elif on_conflict == 'error':
        Candle.insert_many(candles_list).execute()
    else:
        raise Exception(f'Unknown on_conflict value: {on_conflict}')


def fetch_candles_from_db(exchange: str, symbol: str, timeframe: str, start_date: int, finish_date: int) -> tuple:
    res = tuple(
        Candle.select(
            Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
            Candle.volume
        ).where(
            Candle.exchange == exchange,
            Candle.symbol == symbol,
            Candle.timeframe == timeframe,
            Candle.timestamp.between(start_date, finish_date)
        ).order_by(Candle.timestamp.asc()).tuples()
    )

    return res
